"""

 2020 (c) piteren

    Parameters Save

        - saves/loads DNA to/from folder (serializes object)
        - implements GX for saved

"""

import os
import shutil
from typing import Optional, List
import warnings

from pypaq.lipytools.little_methods import stamp
from pypaq.lipytools.files import r_pickle, w_pickle, prep_folder
from pypaq.lipytools.pylogger import get_pylogger, get_hi_child
from pypaq.pms.base_types import POINT
from pypaq.pms.subscriptable import SubGX


class ParaSave(SubGX):

    SAVE_TOPDIR = None
    SAVE_FN_PFX = 'dna' # filename (DNA) prefix
    OBJ_SUFFIX = '.dct' # dna obj(pickle) filename suffix
    TXT_SUFFIX = '.txt' # dna text filename suffix

    def __init__(
            self,
            name:str,
            save_topdir: Optional[str]= SAVE_TOPDIR,    # ParaSave top directory, when not given folders functionality (load/save) is not accessible
            save_fn_pfx: str=           SAVE_FN_PFX,    # ParaSave filename (DNA) prefix
            gxable: bool or None=       None,           # None sets to default - True, it may override default and save by setting it to True/False
            assert_saved=               False,          # for True asserts that ParaSave has been already saved in save_topdir
            lock_managed_params=        False,          # locks _managed_params to only those known while init
            logger=                     None,
            loglevel=                   20,
            **kwargs):

        self.name = name

        if not logger:
            logger = get_pylogger(
                name=       self.name,
                folder=     ParaSave.__full_dir(name=self.name, save_topdir=save_topdir),
                level=      loglevel)
        self.__log = logger

        if assert_saved and not os.path.isfile(ParaSave.__obj_fn(name, save_topdir, save_fn_pfx)):
            ex_msg = f'ParaSave {self.name} does not exist!'
            self.__log.error(ex_msg)
            raise Exception(ex_msg)

        self.__log.info(f'*** ParaSave *** name: {self.name} initializes..')

        self.save_topdir = save_topdir
        self.save_fn_pfx = save_fn_pfx
        self.gxable = True
        self.parents = [] # list of parents names TODO: what is this for?

        dna_folder = ParaSave.load_dna(
            name=           self.name,
            save_topdir=    self.save_topdir,
            save_fn_pfx=    self.save_fn_pfx)
        self.update(dna_folder) # 1. update with params from folder
        self.update(kwargs)     # 3. update with params given by user
        if gxable is not None: self.gxable = gxable  # if user forces it to be True/False (no matter what saved)

        # _managed_params allows to lock managed params only to those given here
        self._managed_params: Optional[List[str]] = None
        if lock_managed_params: self._managed_params = self.get_managed_params()

        SubGX.__init__(self, logger=get_hi_child(self.__log), **self.get_point())


    def get_managed_params(self) -> List[str]:
        if self._managed_params is not None: return self._managed_params
        return SubGX.get_managed_params(self)

    # ************************************************************************************************ folder management

    @staticmethod
    def __full_dir(
            name: str,
            save_topdir: str):
        return f'{save_topdir}/{name}'

    @classmethod
    def __obj_fn(
            cls,
            name: str,
            save_topdir: str,
            save_fn_pfx: str):
        return f'{cls.__full_dir(name, save_topdir)}/{save_fn_pfx}{cls.OBJ_SUFFIX}'

    @classmethod
    def __txt_fn(
            cls,
            name: str,
            save_topdir: str,
            save_fn_pfx: str):
        return f'{cls.__full_dir(name, save_topdir)}/{save_fn_pfx}{cls.TXT_SUFFIX}'

    # loads ParaSave DNA from folder
    @classmethod
    def load_dna(
            cls,
            name: str,
            save_topdir: Optional[str]= None,
            save_fn_pfx: Optional[str]= None) -> POINT:
        obj_FN = ParaSave.__obj_fn(
            name=           name,
            save_topdir=    save_topdir or cls.SAVE_TOPDIR,
            save_fn_pfx=    save_fn_pfx or cls.SAVE_FN_PFX)
        if os.path.isfile(obj_FN):
            dna: POINT = r_pickle(obj_FN)
            return dna
        return {}

    # saves ParaSave DNA to folder (with preview in txt)
    def save_dna(self):

        assert self.save_topdir, 'ERR: cannot save ParaSave, if save directory was not given, aborting!'

        obj_FN = ParaSave.__obj_fn(
            name=           self.name,
            save_topdir=    self.save_topdir,
            save_fn_pfx=    self.save_fn_pfx)
        txt_FN = ParaSave.__txt_fn(
            name=           self.name,
            save_topdir=    self.save_topdir,
            save_fn_pfx=    self.save_fn_pfx)

        prep_folder(ParaSave.__full_dir(
            name=           self.name,
            save_topdir=    self.save_topdir))

        if os.path.isfile(obj_FN): shutil.copy(obj_FN, f'{obj_FN}_OLD')
        if os.path.isfile(txt_FN): shutil.copy(txt_FN, f'{txt_FN}_OLD')

        dna: POINT = self.get_point()
        w_pickle(dna, obj_FN)
        with open(txt_FN, 'w') as file:
            s = f' *** ParaSave DNA saved: {stamp(year=True, letters=None)} ***\n'
            s += ParaSave.dict_2str(dna)
            file.write(s)

    # loads, next overrides parameters from given kwargs and saves new ParaSave DNA
    @classmethod
    def oversave(
            cls,
            name: str,
            save_topdir: Optional[str]= None,
            save_fn_pfx: Optional[str]= None,
            **kwargs):
        psc = ParaSave(
            name=           name,
            save_topdir=    save_topdir or cls.SAVE_TOPDIR,
            save_fn_pfx=    save_fn_pfx or cls.SAVE_FN_PFX,
            loglevel=       30)
        psc.update(kwargs)
        psc.save_dna()

    # copies saved ParaSave DNA from one folder to another
    @classmethod
    def copy_saved_dna(
            cls,
            name_src: str,
            name_trg: str,
            save_topdir_src: Optional[str]= None,
            save_topdir_trg: Optional[str]= None,
            save_fn_pfx: Optional[str]=     None):

        if not save_topdir_src: save_topdir_src = cls.SAVE_TOPDIR
        if not save_fn_pfx: save_fn_pfx = cls.SAVE_FN_PFX

        ps_src = ParaSave(
            name=           name_src,
            save_topdir=    save_topdir_src,
            save_fn_pfx=    save_fn_pfx,
            assert_saved=   True,
            loglevel=       30)

        ps_trg = ParaSave(
            name=           name_trg,
            save_topdir=    save_topdir_trg or save_topdir_src,
            save_fn_pfx=    save_fn_pfx,
            loglevel=       30)

        dna_src = ps_src.get_point()
        for k in ['name','save_topdir']: dna_src.pop(k)
        ps_trg.update(dna_src)
        ps_trg.save_dna()

    # performs GX on saved ParaSave DNA (without even building child objects)
    @classmethod
    def gx_saved_dna(
            cls,
            name_parent_main: str,
            name_parent_scnd: Optional[str],                        # if not given makes GX only with main parent
            name_child: str,
            save_topdir_parent_main: Optional[str]= None,   # ParaSave top directory
            save_topdir_parent_scnd: Optional[str]= None,   # ParaSave top directory of parent scnd
            save_topdir_child: Optional[str]=       None,   # ParaSave top directory of child
            save_fn_pfx: Optional[str]=             None,   # ParaSave dna filename prefix
    ) -> None:

        if not save_topdir_parent_main: save_topdir_parent_main = cls.SAVE_TOPDIR
        if not save_topdir_parent_scnd: save_topdir_parent_scnd = save_topdir_parent_main
        if not save_topdir_child: save_topdir_child = save_topdir_parent_main
        if not save_fn_pfx: save_fn_pfx = cls.SAVE_FN_PFX

        pm = ParaSave(
            name=           name_parent_main,
            save_topdir=    save_topdir_parent_main,
            save_fn_pfx=    save_fn_pfx,
            assert_saved=   True,
            loglevel=       30)

        ps = ParaSave(
            name=           name_parent_scnd,
            save_topdir=    save_topdir_parent_scnd,
            save_fn_pfx=    save_fn_pfx,
            loglevel=       30) if name_parent_scnd else None

        not_gxable_parents = []
        if not pm.gxable: not_gxable_parents.append(pm)
        if ps and not ps.gxable: not_gxable_parents.append(ps)
        if not_gxable_parents: warnings.warn('There are not gxable parents, cannot GX!')
        else:
            # make pm a child and save
            dna_child = SubGX.gx_dna(
                parent_main=    pm,
                parent_scnd=    ps,
                name_child=     name_child)
            pm.update(dna_child)
            pm['_save_topdir'] = save_topdir_child
            pm.parents = [name_parent_main]
            if name_parent_scnd: pm.parents.append(name_parent_scnd)
            pm.save_dna()

    # adds gxable check
    @staticmethod
    def gx_dna(
            parent_main: "ParaSave",
            parent_scnd: Optional["ParaSave"]=  None,
            name_child: Optional[str]=          None,
            prob_mix=                           0.5,
            prob_noise=                         0.3,
            noise_scale=                        0.1,
            prob_axis=                          0.1,
            prob_diff_axis=                     0.3) -> POINT:
        not_gxable_parents = []
        if not parent_main.gxable: not_gxable_parents.append(parent_main)
        if parent_scnd and not parent_scnd.gxable: not_gxable_parents.append(parent_scnd)
        if not_gxable_parents: warnings.warn('There are not-gxable parents, cannot GX!')
        else:
            return SubGX.gx_dna(
                parent_main=    parent_main,
                parent_scnd=    parent_scnd,
                name_child=     name_child,
                prob_mix=       prob_mix,
                prob_noise=     prob_noise,
                noise_scale=    noise_scale,
                prob_axis=      prob_axis,
                prob_diff_axis= prob_diff_axis)

    # returns difference in POINT of saved VS self
    def _get_diff_saved(self) -> POINT:
        dna_current = self.get_point()
        dna_saved = ParaSave.load_dna(
            name=           self.name,
            save_topdir=    self.save_topdir,
            save_fn_pfx=    self.save_fn_pfx)
        return {k: dna_saved[k] for k in dna_saved if dna_saved[k]!=dna_current[k]}

    # reloads and applies new values of parameters from save (allows to apply new settings after GX of saved)
    def reload(self) -> POINT:
        warnings.warn(f'NOT IMPLEMENTED reload() for {type(self)}(ParaSave)!')
        return self._get_diff_saved()

    # returns nice string of given dict (mostly for .txt preview save)
    @staticmethod
    def dict_2str(d: dict) -> str:
        if d:
            s = ''
            max_len_sk = max([len(k) for k in d.keys()])
            for k, v in sorted(d.items()): s += f'{str(k):{max_len_sk}s} : {str(v)}\n'
            return s[:-1]
        return '--empty dna--'

    def __str__(self):
        s = f'(ParaSave) name: {self.name}, family: {self.family}\n'
        s += self.dict_2str(self.get_point())
        return s


"""

 2022 (c) piteren

    Subscriptable - class of objects which parameters may be accessed in dict-like style [], but are not dict type
        - protected fields may also be accessed with [], but rather should not

    SubGX

        Class of objects that may perform GX on DNA

        DNA is a POINT in context of class __init__, type(DNA) == type(POINT),
        it is a complete set of kwargs needed to perform init of a class instance.

        DNA of child comes from parent fields and those sampled with GX algorithm.

        GX of one parent is a procedure of sampling child from the space of parent with GXA parameters.
"""

from copy import deepcopy
from typing import List, Optional, Union, Dict

from pypaq.lipytools.pylogger import get_pylogger, get_hi_child
from pypaq.textools.text_metrics import lev_dist
from pypaq.pms.base_types import POINT, PSDD
from pypaq.pms.paspa import PaSpa


class Subscriptable:

    def __init__(self, logger=None):
        self.__log = logger or get_pylogger()

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __contains__(self, key):
        return key in vars(self)

    # ********************************************************************************************************** getters

    # returns list of all object fields
    def get_all_fields(self, protected=False) -> List[str]:
        fields = [pm for pm in vars(self)]
        if not protected: fields = [f for f in fields if not f.startswith('_')]
        return fields

    # prepares list of params managed by Subscriptable (all but private and protected)
    def get_managed_params(self) -> List[str]:
        return self.get_all_fields(protected=False)

    # returns dict of params managed by Subscriptable (POINT of ALL object fields)
    def get_point(self) -> POINT:
        return {k: self[k] for k in self.get_managed_params()}

    # *************************************************************************************************** update & check

    # update in dict-like style
    def update(self, dct:dict):
        for key in dct:
            self.__setitem__(key, dct[key])

    # checks for self.params + params for similarity, returns True if already got similar
    def check_params_sim(
            self,
            params: Optional[Union[Dict,List]]= None,
            lev_dist_diff: int=                 1):

        if params is None: params = []
        paramsL = params if type(params) is list else list(params.keys())
        self_paramsL = list(self.get_point().keys())
        diff_paramsL = [par for par in paramsL if par not in self_paramsL]

        all_params = sorted(list(set([p.lower() for p in self_paramsL + diff_paramsL])))

        #found_any = False
        found = set()
        for pa in all_params:
            for pb in all_params:
                if pa != pb:
                    levD = lev_dist(pa,pb)
                    if levD <= lev_dist_diff:
                        found.add(tuple(sorted([pa, pb])))

        if found:
            self.__log.warning('Subscriptable was asked to check for params similarity and found:')
            for pa,pb in found: self.__log.warning(f'> params \'{pa}\' and \'{pb}\' are too CLOSE !!!')

        return len(found) > 0

    def __str__(self):
        s = f'\n(Subscriptable):\n'
        pms = self.get_point()
        for k in sorted(pms.keys()):
            p = f'param: {k}'
            v = f'value: {pms[k]}'
            s += f'{p:30s} {v:30s}\n'
        return s[:-1]

# adds GX on DNA to Subscriptable
class SubGX(Subscriptable):

    def __init__(
            self,
            name: str,
            family: Optional[str]=  None, # family of GXable
            psdd: Optional[PSDD]=   None, # PSDD of GXable
            logger=                 None,
            **kwargs):

        self.__log = logger or get_pylogger()

        Subscriptable.__init__(self, logger=get_hi_child(self.__log))

        self.name = name
        self.family = family
        self.psdd = psdd or {}
        self.update(kwargs)
        # INFO: all keys of self.psdd should be present in self
        #  - it is not checked now (while init) since self may be updated even after init, IT IS checked while GX
        self.__log.info(f'*** Subscriptable *** name: {self.name} initialized, family: {self.family}, psdd: {self.psdd}')

    # returns self POINT limited to axes included in self.psdd
    def get_gxable_point(self) -> POINT:
        point_gxable = {k: self[k] for k in self.psdd}
        return point_gxable

    # checks for compatibility of families stored in given DNA, True when are equal or when at least one family is None
    @staticmethod
    def families_compatible(
            parent_main: "SubGX",
            parent_scnd: Optional["SubGX"]= None) -> bool:

        dna_main = parent_main.get_point()
        dna_scnd = parent_scnd.get_point() if parent_scnd else None

        if dna_scnd is None: return True

        if not dna_scnd: dna_scnd = {}
        fmm = dna_main.get('family', None)
        fms = dna_scnd.get('family', None)
        if fmm is not None and fms is not None and fmm != fms: return False
        return True

    # GX on DNA (POINTs)
    @staticmethod
    def gx_dna(
            parent_main: "SubGX",
            parent_scnd: Optional["SubGX"]= None,
            name_child: Optional[str]=      None,
            prob_mix=                       0.5,
            prob_noise=                     0.3,
            noise_scale=                    0.1,
            prob_axis=                      0.1,
            prob_diff_axis=                 0.3) -> POINT:

        assert SubGX.families_compatible(parent_main, parent_scnd), 'ERR: families not compatible'

        dna_main = parent_main.get_point()
        dna_scnd = parent_scnd.get_point() if parent_scnd else None

        psdd_main = dna_main.get('psdd', {})
        psdd_scnd = dna_scnd.get('psdd', {}) if dna_scnd is not None else None
        psdd_merged = PaSpa.merge_psdd(
                psdd_a= psdd_main,
                psdd_b= psdd_scnd) if psdd_scnd is not None else psdd_main

        paspa_merged = PaSpa(psdd=psdd_merged)

        paspa_axes_not_in_parent = [a for a in paspa_merged.axes if a not in dna_main]
        assert not paspa_axes_not_in_parent, f'ERR: paspa axes not in parent_main: {paspa_axes_not_in_parent}'

        # sub-points limited to axes of psdd_merged (PaSpa.sample_point_GX() does not accept other axes..)
        dna_main_psdd = {k: dna_main[k] for k in psdd_merged}
        dna_scnd_psdd = {k: dna_scnd[k] for k in psdd_merged} if dna_scnd is not None else None

        point_gx = paspa_merged.sample_point_GX(
            point_main=     dna_main_psdd,
            point_scnd=     dna_scnd_psdd,
            prob_mix=       prob_mix,
            prob_noise=     prob_noise,
            noise_scale=    noise_scale,
            prob_axis=      prob_axis,
            prob_diff_axis= prob_diff_axis)

        # dna_child is based on parent_main with updated values of point_gx
        dna_child: POINT = deepcopy(dna_main)
        dna_child.update(point_gx)
        dna_child['psdd'] = psdd_merged # update psdd to psdd_merged

        name_merged = f'{parent_main.name}+{parent_scnd.name}_(gxp)' if parent_scnd else f'{parent_main.name}_(sgxp)'
        dna_child['name'] = name_child or name_merged

        return dna_child

    def __str__(self):
        s = f'\n(Subscriptable) name: {self.name}, family: {self.family}\n'
        pms = self.get_point()
        for k in sorted(pms.keys()):
            p = f'param: {k}'
            v = f'value: {pms[k]}'
            r = f'ranges:{self.psdd[k]}' if k in self.psdd else ''
            s += f'{p:30s} {v:30s} {r}\n'
        return s[:-1]


import random
from typing import Sized, List, Tuple, Optional, Iterable
import warnings

from pypaq.hpmser.helpers import _str_weights
from pypaq.lipytools.files import r_pickle, w_pickle
from pypaq.lipytools.plots import three_dim
from pypaq.pms.paspa import PaSpa
from pypaq.pms.base_types import POINT, point_str


# Search Results List [SeRes] with some methods
class SRL(Sized):

    # single Search Result
    class SeRes:

        def __init__(
                self,
                point: POINT,
                score=      None):
            self.id = None              # to be updated, id of SeRes, set by SRL: id = len(SRL)
            self.point = point
            self.score = score
            self.smooth_score = None    # to be updated

        def __str__(self): return f'SeRes: id:{self.id}, point:{self.point}, score:{self.score}, smooth_score{self.smooth_score}'

    def __init__(
            self,
            paspa: Optional[PaSpa]=     None,   # parameters space of this SRL
            name: str=                  'SRL',
            np_smooth: int=             3,      # (NPS) Number of Points taken into account while calculating smooth score - default / starting value
            plot_axes: list=            None,   # list with axes names (max 3), eg: ['drop_a','drop_b','loss']
            verb=                       0):

        self.verb = verb
        self.name = name
        if self.verb>0: print(f'\n*** SRL {self.name} initializing')

        self.paspa = paspa

        self.__np_smooth = np_smooth        # current np_smooth of SeRes
        self.plot_axes = plot_axes


        self.__srL: List[SRL.SeRes] = []    # sorted periodically by smooth_score
        self.__smoothed_and_sorted = True   # flag (state) - keeps info about SRL that is smoothed & sorted
        self.__distances = []               # distances cache (by SeRes id)
        self.__scores = []                  # scores cache (by SeRes id)
        self.__avg_dst = 1                  # average distance of SRL for self.__np_smooth
        self.prec = 8                       # print precision, will be updated while adding new points

    # ****************************************************************************************************** load & save

    def __get_srl_path(self, save_dir: str) -> str: return f'{save_dir}/{self.name}.srl'

    def __get_srl_backup_path(self, save_dir: str) -> str: return f'{save_dir}/{self.name}.srl.backup'

    # loads (alternatively from backup)
    def load(self, save_dir :str):

        if self.verb > 0: print(f' > SRL {self.name} loading form {save_dir}..')

        try:
            obj = r_pickle(self.__get_srl_path(save_dir))
        except Exception as e:
            print(f' SRL {self.name} got exception: {str(e)} while loading, using backup file')
            obj = r_pickle(self.__get_srl_backup_path(save_dir))

        self.paspa =                    obj.paspa
        self.__np_smooth =              obj.__np_smooth
        self.plot_axes =                obj.plot_axes

        self.__srL =                    obj.__srL
        self.__smoothed_and_sorted =    obj.__smoothed_and_sorted
        self.__distances =              obj.__distances
        self.__scores =                 obj.__scores
        self.__avg_dst =                obj.__avg_dst
        self.prec=                      obj.prec

        if self.verb>0: print(f' > SRL loaded {len(self.__srL)} results')

    # saves with backup
    def save(self, folder :str):

        # backup copy previous
        old_res = r_pickle(self.__get_srl_path(folder))
        if old_res: w_pickle(old_res, self.__get_srl_backup_path(folder))

        w_pickle(self, self.__get_srl_path(folder))
        self.plot(folder=folder)

    # ************************************************************************************************ getters & setters

    # returns top SeRes (max smooth_score)
    def get_top_SR(self) -> SeRes or None:
        if self.__srL:
            if not self.__smoothed_and_sorted:
                warnings.warn('SRL asked to return top SR while SRL is not smoothed_and_sorted - have to sort!')
                self.smooth_and_sort()
            return self.__srL[0]
        return None

    def get_SR(self, id: int) -> SeRes or None:
        for sr in self.__srL:
            if sr.id == id: return sr
        return None

    # returns distance between two points, if points are SeRes type uses cached distance
    def get_distance(self,
            pa: POINT or SeRes,
            pb: POINT or SeRes) -> float:
        if type(pa) is SRL.SeRes and type(pb) is SRL.SeRes: return self.__distances[pa.id][pb.id]
        if type(pa) is SRL.SeRes: pa = pa.point
        if type(pb) is SRL.SeRes: pb = pb.point
        return self.paspa.distance(pa, pb)

    def get_avg_dst(self): return self.__avg_dst

    # max of min distances of SRL: max(min_distance)
    def get_mom_dst(self): return max([min(d[1:]) if d[1:] else 0 for d in self.__distances])

    # returns sample with policy and estimated score
    def get_opt_sample(self,
            prob_opt,   # probability of optimized sample
            n_opt,      # number of optimized samples
            prob_top,   # probability of sample from area of top
            n_top,      # number of top samples
            avg_dst     # distance for sample from area of top
    ) -> Tuple[POINT,float]:

        prob_rnd = 1 - prob_opt - prob_top
        if random.random() < prob_rnd or len(self.__srL) < 10: sample = self.paspa.sample_point_GX() # one random point
        else:
            if random.random() < prob_opt/(prob_top+prob_opt): points = [self.paspa.sample_point_GX() for _ in range(n_opt + 1)] # some random points ...last for reference
            # top points
            else:
                n_top += 1 # last for reference
                if n_top > len(self.__srL): n_top = len(self.__srL)
                points = [self.paspa.sample_point_GX(
                    point_main=     self.__srL[ix].point,
                    noise_scale=    avg_dst) for ix in range(n_top)] # top points

            scores = [self.smooth_point(p)[0] for p in points]

            all_pw = list(zip(points, scores))
            all_pw.sort(key=lambda x: x[1], reverse=True)
            maxs = all_pw[0][1]
            subs = all_pw.pop(-1)[1]
            mins = all_pw[-1][1]

            all_p, all_w = zip(*all_pw)
            all_w = [w - subs for w in all_w]
            all_p = list(all_p)
            sample = random.choices(all_p, weights=all_w, k=1)[0]
            pf = f'.{self.prec}f'
            print(f'   % sampled #{all_p.index(sample)}/{len(all_p)} from: {maxs:{pf}}-{mins:{pf}} {_str_weights(all_w, float_prec=self.prec)}')

        est_score, _, _ =  self.smooth_point(sample)

        return sample, est_score

    # returns sample chosen with policy and its estimated score
    def get_opt_sample_GX(
            self,
            prob_opt,   # probability of optimized sample
            n_opt,      # number of optimized samples
            prob_top,   # probability of sample from area of top
            n_top,      # number of top samples
            avg_dst     # distance for sample from area of top
    ) -> Tuple[POINT,float]:

        if random.random() < 1-prob_opt-prob_top or len(self.__srL) < 10:
            sample = self.paspa.sample_point_GX() # one random point
        else:
            # choice from better half of random points
            if random.random() < prob_opt/(prob_top+prob_opt):
                points = [self.paspa.sample_point_GX() for _ in range(2 * n_opt)]   # 2*some random points
                scores = [self.smooth_point(p)[0] for p in points]                  # their scores
                all_pw = list(zip(points, scores))
                all_pw.sort(key=lambda x: x[1], reverse=True)                       # sorted
                all_pw = all_pw[:len(all_pw)//2]                                    # take better half
                maxs = all_pw[0][1]
                subs = all_pw.pop(-1)[1]
                mins = all_pw[-1][1]
                all_p, all_w = zip(*all_pw)
                all_w = [w - subs for w in all_w]
                all_p = list(all_p)
                sample = random.choices(all_p, weights=all_w, k=1)[0]
                pf = f'.{self.prec}f'
                print(f'   % sampled #{all_p.index(sample)}/{len(all_p)} from: {maxs:{pf}}-{mins:{pf}} {_str_weights(all_w, float_prec=self.prec)}')
            # GX from top points
            else:
                n_top += 1 # last for reference
                if n_top > len(self.__srL): n_top = len(self.__srL)
                top_sr = self.__srL[:n_top]                             # top SR
                scores = [p.smooth_score for p in top_sr]               # their scores
                mins = min(scores)
                scores = [s-mins for s in scores]                       # reduced scores
                sra, srb = random.choices(top_sr, weights=scores, k=2)  # two SR
                sample = self.paspa.sample_point_GX(
                    point_main=                 sra.point,
                    point_scnd=                 srb.point,
                    noise_scale=                avg_dst)
                pf = f'.{self.prec}f'
                print(f'   % sampled GX from: {sra.smooth_score:{pf}} and {srb.smooth_score:{pf}}')
        est_score, _, _ =  self.smooth_point(sample)
        return sample, est_score

    # returns n closest SeRes to given point
    def __get_n_closest(
            self,
            point: POINT or SeRes,
            n: Optional[int]=   None) -> List[SeRes]:
        if not n: n = self.__np_smooth
        if len(self.__srL) <= n:
            return [] + self.__srL
        else:
            id_dst = \
                list(zip(range(len(self.__srL)), self.__distances[point.id])) \
                    if type(point) is SRL.SeRes else \
                    [(sr.id, self.get_distance(point, sr.point)) for sr in self.__srL]
            id_dst.sort(key=lambda x: x[1]) # sort by distance to this point
            return [self.get_SR(id[0]) for id in id_dst[:n]]

    # sets np_smooth, then smooths and sorts
    def set_np_smooth(self, np_smooth: int):
        if np_smooth != self.__np_smooth:
            self.__np_smooth = np_smooth
            self.smooth_and_sort()

    # adds new result, caches distances, smooths & sorts
    def add_result(self,
            point: POINT,
            score: float,
            force_no_update=    False # aborts calculation of smooth score and sorting
    ) -> SeRes:

        sr = SRL.SeRes(point,score)
        sr.id = len(self.__srL)

        # update cached distances
        sr_dist = []
        id_point = [(s.id,s.point) for s in self.__srL]
        id_point.sort(key= lambda x:x[0]) # sort by id
        for id,point in id_point:
            d = self.paspa.distance(sr.point,point)
            sr_dist.append(d)
            self.__distances[id].append(d)
        sr_dist.append(0)

        self.__distances.append(sr_dist)

        self.__scores.append(score) # update cached score

        self.__srL.append(sr) # add SeRes

        if force_no_update:
            sr.smooth_score = sr.score
            self.__smoothed_and_sorted = False
        else: self.smooth_and_sort()

        if score > 0.01: self.prec = 4

        return sr

    # returns (smooth_score, average_distance, all scores sorted by distance) for given point or SeRes
    def smooth_point(self,
            point: POINT or SeRes,
            np_smooth: Optional[int]=   None # used if given, otherwise self.__np_smooth
    ) -> Tuple[float, float, List[float]]:

        # case: no points in srL
        smooth_score_np = 0     # smooth score for self.__np_smooth
        avg_dst_np =      1     # average distance for np_smooth
        all_scores =     [0]    # np scores sorted by distance

        if not np_smooth: np_smooth = self.__np_smooth
        if self.__srL:
            score_dst = \
                list(zip(self.__scores, self.__distances[point.id])) \
                    if type(point) is SRL.SeRes else \
                    [(sr.score, self.get_distance(point, sr.point)) for sr in self.__srL]
            score_dst.sort(key=lambda x: x[1])  # sort by distance to this point
            score_dst_np = score_dst[:np_smooth + 1] # trim to np_smooth points (+1 point for reference)

            # case of one/two points in score_dst_np
            if len(score_dst_np) < 3:
                smooth_score_np = score_dst_np[0][0]  # closest point score
                all_scores = [score_dst_np[0][0]]
            else:
                all_scores, all_dst = zip(*score_dst_np)  # scores, distances

                max_dst = all_dst[-1]  # distance of last(reference) point

                # remove last (reference)
                all_dst = all_dst[:-1]
                all_scores = all_scores[:-1]

                # set weights for scores
                weights = []
                if max_dst: weights = [(max_dst-d)/max_dst for d in all_dst]  # try with distance based weights <1;0>
                if sum(weights) == 0: weights = [1] * len(all_dst)  # naive baseline / equal for case: (max_dst == 0) or (max_dst-d == 0)

                wall_scores = [all_scores[ix]*weights[ix] for ix in range(len(all_scores))]
                smooth_score_np = sum(wall_scores) / sum(weights) # weighted score
                avg_dst_np = sum(all_dst) / len(all_dst)

        return smooth_score_np, avg_dst_np, all_scores

    # smooths self.__srL and sorts by SeRes.smooth_score
    def smooth_and_sort(self):

        if self.__srL:
            avg_dst = []
            for sr in self.__srL:
                sr.smooth_score, ad, _ = self.smooth_point(sr)
                avg_dst.append(ad)
            self.__avg_dst = sum(avg_dst)/len(avg_dst)

            self.__srL.sort(key=lambda x: x.smooth_score, reverse=True)

        self.__smoothed_and_sorted = True

    def print_distances(self):
        for dl in self.__distances:
            s = ''
            for d in dl: s += f'{d:.2f} '
            print(s)

    def nice_str(
            self,
            n_top=                  20,
            top_nps: Iterable[int]= (3,5,9),
            all_nps: Optional[int]= 3):

        pf = f'.{self.prec}f'

        re_str = ''
        if all_nps: re_str += f'Search run {self.name}, {len(self.__srL)} results:\n\n{self.paspa}\n\n'

        if len(self.__srL) < n_top: n_top = len(self.__srL)
        orig_nps = self.__np_smooth
        top_nps = list(top_nps)
        for nps in top_nps:
            self.set_np_smooth(nps)

            re_str += f'TOP{n_top} results for NPS {nps} (avg_dst:{self.__avg_dst:.3f}):'
            if top_nps.index(nps) == 0: re_str += ' -- id smooth [local] [max-min] avg_dst {params..}\n'
            else: re_str += '\n'

            for srIX in range(n_top):
                sr = self.__srL[srIX]
                ss_np, avg_dst, all_scores = self.smooth_point(sr)
                re_str += f'{sr.id:4d} {ss_np:{pf}} [{sr.score:{pf}}] [{max(all_scores):{pf}}-{min(all_scores):{pf}}] {avg_dst:.3f} {point_str(sr.point)}\n'

        self.set_np_smooth(orig_nps)
        top_sr_nps = self.get_top_SR()
        n_closest = self.__get_n_closest(top_sr_nps, n=self.__np_smooth)
        re_str += f'{self.__np_smooth} closest points to TOP (NPS {self.__np_smooth}):\n'
        for sr in n_closest: re_str += f'{sr.id:4d} [{sr.score:{pf}}] {point_str(sr.point)}\n'

        if all_nps and len(self.__srL) > n_top:
            self.set_np_smooth(all_nps)
            re_str += f'\nALL results for NPS {all_nps} (avg_dst:{self.__avg_dst:.3f}):\n'
            for sr in self.__srL:
                ss_np, avg_dst, all_scores = self.smooth_point(sr)
                re_str += f'{sr.id:4d} {ss_np:{pf}} [{sr.score:{pf}}] [{max(all_scores):{pf}}-{min(all_scores):{pf}}] {avg_dst:.3f} {point_str(sr.point)}\n'

        self.set_np_smooth(orig_nps)
        return re_str

    # 3D plot
    def plot(
            self,
            smooth_score=   True,   # for True color_"axis" == smooth_score, else score
            folder: str=    None):

        columns = sorted(list(self.__srL[0].point.keys()))[:3] if not self.plot_axes else [] + self.plot_axes
        valLL = [[res.point[key] for key in columns] for res in self.__srL]

        # eventually add score
        if len(columns) < 3:
            valLL = [vl + [res.score] for vl,res in zip(valLL,self.__srL)]
            columns += ['score']

        # eventually add smooth_score (for one real axis)
        if len(columns) < 3:
            valLL = [vl + [res.smooth_score] for vl,res in zip(valLL,self.__srL)]
            columns += ['smooth_score']

        # add color "axis" data
        columns += ['smooth_score' if smooth_score else 'score']
        valLL = [valL + [res.smooth_score if smooth_score else res.score] for valL, res in zip(valLL, self.__srL)]

        three_dim(
            xyz=        valLL,
            name=       self.name,
            x_name=     columns[0],
            y_name=     columns[1],
            z_name=     columns[2],
            val_name=   columns[3],
            save_FD=    folder)

    def __len__(self): return len(self.__srL)


"""

 2021 (c) piteren

    PG_Actor - PolicyGradients TrainableActor, NN based

    TODO: implement parallel training, in batches (many envys)

"""

from abc import ABC, abstractmethod
import numpy as np
from typing import Optional, Callable, Union, List

from pypaq.lipytools.little_methods import stamp
from pypaq.R4C.actor import TrainableActor
from pypaq.R4C.envy import FiniteActionsRLEnvy
from pypaq.comoneural.nnwrap import NNWrap


class PGActor(TrainableActor, ABC):

    def __init__(
            self,
            envy: FiniteActionsRLEnvy,
            nnwrap: type(NNWrap),
            nngraph: Optional[Union[Callable, type]]=   None,
            seed: int=                                  123,
            **kwargs):

        TrainableActor.__init__(
            self,
            envy=   envy,
            **kwargs)
        self._envy = envy  # to update type (for pycharm only)

        np.random.seed(seed)

        # some overrides and updates
        kwargs['name'] = self.name
        kwargs['name_timestamp'] = False                # name timestamp is driven by TrainableActor (with self.name)
        if 'logger' in kwargs: kwargs.pop('logger')     # NNWrap will always create own logger (since then it is not given) with optionally given level
        kwargs['num_actions'] = self._envy.num_actions()
        kwargs['observation_width'] = self._get_observation_vec(self._envy.get_observation()).shape[-1]

        self.nnw: NNWrap = nnwrap(nngraph=nngraph, seed=seed, **kwargs)

        self._rlog.info('*** PG_Actor *** (NN based) initialized')
        self._rlog.info(f'> NNWrap: {nnwrap.__name__}')

    # vectorization of observations batch, may be overridden with more optimal custom implementation
    def _get_observation_vec_batch(self, observations: List[object]) -> np.ndarray:
        return np.array([self._get_observation_vec(v) for v in observations])

    @abstractmethod
    def get_policy_probs(self, observation: object) -> np.ndarray: pass

    # baseline, may be overridden with more optimal custom implementation
    def get_policy_probs_batch(self, observations: List[object]) -> np.ndarray:
        return np.array([self.get_policy_probs(o) for o in observations])

    # gets policy action based on policy (action) probs
    def get_policy_action(self, observation: object, sampled=False) -> int:
        probs = self.get_policy_probs(observation)
        if sampled: action = np.random.choice(self._envy.num_actions(), p=probs)
        else:       action = np.argmax(probs)
        return int(action)

    def get_policy_action_batch(self, observations: List[object], sampled=False) -> np.ndarray:
        probs = self.get_policy_probs_batch(observations)
        if sampled: actions = np.random.choice(self._envy.num_actions(), size=probs.shape[-1], p=probs)
        else:       actions = np.argmax(probs, axis=-1)
        return actions

    def _get_save_topdir(self) -> str:
        return self.nnw['save_topdir']

    def save(self):
        self.nnw.save()

    def __str__(self):
        return self.nnw.__str__()


"""

 2018 (c) piteren

    some little methods (but frequently used) for Python

"""

from collections import OrderedDict
import inspect
import random
import string
import time
from typing import Dict, List, Callable, Any, Optional


# prepares function parameters dictionary
def get_params(function: Callable):
    params_dict = {'without_defaults':[], 'with_defaults':OrderedDict()}
    if function:
        specs = inspect.getfullargspec(function)
        params = specs.args
        if not params: params = []
        vals = specs.defaults
        if not vals: vals = ()

        while len(params) > len(vals):
            params_dict['without_defaults'].append(params.pop(0))

        params_dict['with_defaults'] = {k: v for k,v in zip(params,vals)}

    return params_dict

# prepares func sub-DNA given full DNA (wider)
def get_func_dna(
        func: Optional[Callable],
        dna: Dict,
        remove_self= True # removes self in case of methods (class)
) -> dict:
    if func is None: return {}
    pms = get_params(func)
    valid_keys = pms['without_defaults'] + list(pms['with_defaults'].keys())
    if remove_self and 'self' in valid_keys: valid_keys.remove('self')
    func_dna = {k: dna[k] for k in dna if k in valid_keys} # filter to get only params accepted by func
    return func_dna

# short(compressed) scientific notation for floats
def short_scin(
        fl: float,
        precision:int=  1):
    sh = f'{fl:.{precision}E}'
    sh = sh.replace('+0','')
    sh = sh.replace('+','')
    sh = sh.replace('-0','-')
    sh = sh.replace('E','e')
    return sh

# returns sting from float, always of given width
def float_to_str(
        num: float,
        width: int= 7):
    if width < 5: width = 5
    scientific_decimals = width-6 if width>6 else 0
    ff = f'{num:.{scientific_decimals}E}'
    if 1000 > num > 0.0001: ff = str(num)[:width]
    if len(ff)<width: ff += '0'*(width-len(ff))
    return ff

# returns timestamp string
def stamp(
        year=                   False,
        date=                   True,
        letters: Optional[int]= 3):
    random.seed(time.time())
    if date:
        if year: stp = time.strftime('%y%m%d_%H%M')
        else:    stp = time.strftime('%m%d_%H%M')
    else:        stp = ''
    if letters:
        if date: stp += '_'
        stp += ''.join([random.choice(string.ascii_letters) for _ in range(letters)])
    return stp

# returns nice string of given list
def list_str(ls: List[Any], limit:Optional[int]=200):
    lstr = [str(e) for e in ls]
    lstr = '; '.join(lstr)
    if limit: lstr = lstr[:limit]
    return lstr

# prints nested dict
def print_nested_dict(dc: dict, ind_scale=2, line_limit=200):

    tpD = {
        dict:   'D',
        list:   'L',
        tuple:  'T',
        str:    'S'}

    def __prn_root(root: dict, ind, ind_scale=2, line_limit=line_limit):

        spacer = ' ' * ind * ind_scale
        for k in sorted(list(root.keys())):
            tp = tpD.get(type(root[k]),'O')
            ln = len(root[k]) if tp in tpD.values() else ''

            exmpl = ''
            if tp!='D':
                exmpl = str(root[k])
                if line_limit:
                    if len(exmpl)>line_limit: exmpl = f'{exmpl[:line_limit]}..'
                exmpl = f' : {exmpl}'

            print(f'{spacer}{k} [{tp}.{ln}]{exmpl}')

            if type(root[k]) is dict: __prn_root(root[k],ind+1,ind_scale)

    __prn_root(dc,ind=0,ind_scale=ind_scale)

# prints line over line
def printover(sth):
    print(f'\r{sth}', end='')

# gets folder path from folder or file path

# prepares folder, creates or flushes

# terminal progress bar
def progress_ (
        iteration: float or int,    # current iteration
        total: float or int,        # total iterations
        prefix: str=    '',         # prefix string
        suffix: str=    '',         # suffix string
        length: int=    20,
        fill: str=      '█'):
    prog = iteration / total
    if prog > 1: prog = 1
    filled_length = int(length * prog)
    bar = fill * filled_length + '-' * (length - filled_length)
    printover(f'{prefix} |{bar}| {prog*100:.1f}% {suffix}')
    if prog == 1: print()


"""

 2022 (c) piteren

    PolicyGradients NN Actor, PT based

"""

from abc import ABC
import numpy as np
from typing import Optional

from pypaq.R4C.policy_gradients.pg_actor import PGActor
from pypaq.R4C.policy_gradients.base.pt_based.pg_PT_module import PGModel
from pypaq.torchness.motorch import MOTorch, Module



class PG_PTActor(PGActor, ABC):

    def __init__(
            self,
            name: str=                          'PG_PTActor',
            nngraph: Optional[type(Module)]=    PGModel,
            **kwargs):
        PGActor.__init__(
            self,
            name=       name,
            nnwrap=     MOTorch,
            nngraph=    nngraph,
            **kwargs)

    def get_policy_probs(self, observation: object) -> np.ndarray:
        obs_vec = self._get_observation_vec(observation)
        return self.nnw(obs_vec)['probs'].detach().cpu().numpy()

    # optimized with batch call to NN
    def get_policy_probs_batch(self, observations) -> np.ndarray:
        obs_vecs = self._get_observation_vec_batch(observations)
        return self.nnw(obs_vecs)['probs'].detach().cpu().numpy()

    # updates self NN with batch of data
    def update_with_experience(
            self,
            observations,
            actions,
            dreturns,
            inspect=    False) -> dict:
        obs_vecs = self._get_observation_vec_batch(observations)
        out = self.nnw.backward(obs_vecs, actions, dreturns)
        out.pop('logits')
        if 'probs' in out: out['probs'] = out['probs'].cpu().detach().numpy()
        if 'zeroes' in out: out['zeroes'] = [e.cpu().detach().numpy() for e in out['zeroes']]
        return out


import GPUtil
import os
import platform
from typing import Optional, Union, List

from pypaq.lipytools.pylogger import get_pylogger
from pypaq.mpython.mptools import sys_res_nfo


"""
devices: DevicesParam - (parameter) represents some devices (GPU / CPU)
    # ******************************************************************************************** pypaq representations
    int                 - (int)       single (system) CUDA ID
    -int                - (int)       AVAILABLE CUDA[-int] 
    [] (empty list)     - (list)      all AVAILABLE CUDA
    None                - (NoneType)  single CPU core
    float               - (float)     (0.0;1.0> - factor of system CPU cores
    'all'               - (str)       all CPU cores
    [int,-1,None,'all'] - (list)      list with mix of above, possible repetitions
    # ******************************************************************************** TF1, TF2, PyTorch representations
    '/device:CPU:0'     - (str)       TF1 for CPU
    '/device:GPU:0'     - (str)       TF1 for GPU
    'CPU:0'             - (str)       TF2 for CPU
    'GPU:0'             - (str)       TF2 for GPU
    'cpu:0'             - (str)       PyTorch for CPU
    'cuda:0'            - (str)       PyTorch for GPU
    [str]               - (list)      list with strings of above, possible repetitions   
"""
DevicesParam: Union[int, None, float, str, list] = -1


# returns cuda memory size (system first device)
def get_cuda_mem():
    devs = GPUtil.getGPUs()
    if devs: return devs[0].memoryTotal
    else: return 12000 # safety return for no cuda devices case

# returns list of available GPUs ids
def get_available_cuda_id(max_mem=None) -> list: # None sets automatic, otherwise (0,1.1] (above 1 for all)
    if not max_mem:
        tot_mem = get_cuda_mem()
        if tot_mem < 5000:  max_mem=0.35 # small GPU case, probably system single GPU
        else:               max_mem=0.2
    return GPUtil.getAvailable(limit=20, maxMemory=max_mem)

# prints report of system cuda devices
def report_cuda() -> str:
    rp = 'System CUDA devices:'
    for device in GPUtil.getGPUs():
        rp += f'\n > id: {device.id}, name: {device.name}, MEM: {int(device.memoryUsed)}/{int(device.memoryTotal)} (U/T)'
    return rp

# resolves representation given with DevicesParam into middleform (pypaq: List[Union[int,None]]) or TF / PyTorch supported List[str]
def get_devices(
        devices: DevicesParam=      -1,
        namespace: Optional[str]=   None,
        logger=                     None,
        loglevel=                   20) -> List[Union[int,None,str]]:

    if not logger: logger = get_pylogger(level=loglevel)

    if namespace not in [None,'TF1','TF2','torch']:
        msg = 'Wrong namespace, supported are: None, TF1, TF2 or torch'
        logger.error(msg)
        raise NameError(msg)

    if type(devices) is not list: devices = [devices]  # first convert to list

    cpu_count = sys_res_nfo()['cpu_count']

    # look for available CUDA
    available_cuda_id = []
    if platform.system() == 'Darwin': # OSX
        logger.warning('no GPUs available for OSX, using only CPU')
    else:
        available_cuda_id = get_available_cuda_id()

    if not available_cuda_id:
        logger.debug('no GPUs available, using only CPU')
        num = len(devices)
        if num == 0: num = cpu_count
        devices = [None]*num

    pypaq_devices = []
    if devices == []: pypaq_devices = available_cuda_id
    for d in devices:

        known_device = False

        if type(d) is int:
            if d < 0: pypaq_devices.append(available_cuda_id[d])
            else:     pypaq_devices.append(d)
            known_device = True

        if type(d) is float:
            if d < 0.0: d = 0.0
            if d > 1.0: d = 1.0
            cpu_count_f = round(cpu_count * d)
            if cpu_count_f < 1: cpu_count_f = 1
            pypaq_devices += [None]*cpu_count_f
            known_device = True

        if d == []:
            pypaq_devices += available_cuda_id
            known_device = True

        if type(d) is str:
            if d == 'all':
                pypaq_devices += [None]*cpu_count
                known_device = True
            d_low = d.lower()
            if 'cpu' in d_low:
                pypaq_devices.append(None)
                known_device = True
            if 'gpu' in d_low or 'cuda' in d_low:
                pypaq_devices.append(int(d_low.split(':')[-1]))
                known_device = True

        if d is None:
            pypaq_devices.append(d)
            known_device = True

        if not known_device:
            msg = f'unknown (not valid?) device given: {d}'
            logger.error(msg)
            raise Exception(msg)

    if namespace is None:
        logger.debug(f'get_devices is returning {len(pypaq_devices)} pypaq_devices: {pypaq_devices}')
        return pypaq_devices

    else:

        device_pfx = '/device:'
        if namespace in ['TF2','torch']: device_pfx = ''

        cpu_pfx = 'cpu' if 'TF' not in namespace else 'CPU'
        gpu_pfx = 'cuda' if 'TF' not in namespace else 'GPU'

        lib_devices = []
        for dev in pypaq_devices:
            if type(dev) is int: lib_devices.append(f'{device_pfx}{gpu_pfx}:{dev}')
            if dev is None:      lib_devices.append(f'{device_pfx}{cpu_pfx}:0')

        logger.debug(f'get_devices is returning {len(lib_devices)} lib_devices: {lib_devices}')
        return lib_devices

# masks GPUs from given list of ids or single one
def mask_cuda(ids: Optional[List[int] or int]=  None):
    if ids is None: ids = []
    if type(ids) is int: ids = [ids]
    mask = ''
    for id in ids: mask += f'{int(id)},'
    if len(mask) > 1: mask = mask[:-1]
    os.environ["CUDA_VISIBLE_DEVICES"] = mask

# wraps mask_cuda to hold DevicesParam
def mask_cuda_devices(devices: DevicesParam=-1, logger=None):
    devices = get_devices(devices, namespace=None, logger=logger)
    ids = [d for d in devices if type(d) is int]
    mask_cuda(ids)


"""

 2022 (c) piteren

    baseline PG Module for MOTorch (PyTorch)

"""

import torch

from pypaq.torchness.motorch import Module
from pypaq.torchness.layers import LayDense, zeroes
from pypaq.torchness.base_elements import scaled_cross_entropy


class PGModel(Module):

    def __init__(
            self,
            observation_width=  4,
            num_actions: int=   2,
            hidden_layers=      (20,),
            lay_norm=           False,
            use_scaled_ce=      True,  # experimental Scaled Cross Entropy loss
            seed=               121):

        torch.nn.Module.__init__(self)

        lay_shapeL = []
        next_in = observation_width
        for hl in hidden_layers:
            lay_shapeL.append((next_in,hl))
            next_in = hl

        self.lay_norm = lay_norm

        self.ln = torch.nn.LayerNorm(observation_width) # input layer norm

        self.linL = [LayDense(*shape) for shape in lay_shapeL]
        self.lnL = [torch.nn.LayerNorm(shape[-1]) for shape in lay_shapeL]

        lix = 0
        for lin,ln in zip(self.linL, self.lnL):
            self.add_module(f'lay_lin{lix}', lin)
            self.add_module(f'lay_ln{lix}', ln)
            lix += 1

        self.logits = LayDense(
            in_features=    next_in,
            out_features=   num_actions,
            activation=     None)

        self.use_scaled_ce = use_scaled_ce

    def forward(self, observation) -> dict:

        out = self.ln(observation) if self.lay_norm else observation

        zsL = []
        for lin,ln in zip(self.linL,self.lnL):
            out = lin(out)
            zsL.append(zeroes(out))
            if self.lay_norm: out = ln(out)

        logits = self.logits(out)
        probs = torch.nn.functional.softmax(input=logits, dim=-1)

        return {
            'logits':       logits,
            'probs':        probs,
            'zeroes':       zsL}

    def loss_acc(self, observation, action_taken, dreturn) -> dict:

        out = self(observation)
        logits = out['logits']

        if self.use_scaled_ce:
            actor_ce_scaled = scaled_cross_entropy(
                labels= action_taken,
                scale=  dreturn,
                probs=  out['probs'])
        else:
            actor_ce = torch.nn.functional.cross_entropy(logits, action_taken, reduction='none')
            actor_ce_scaled = actor_ce * dreturn

        out.update({'loss': torch.mean(actor_ce_scaled)})

        return out


"""
 2019 (c) piteren

 NEModel implements NNWrap interface with TensorFlow.

    nngraph for NEModel:
        - should build complete model forward graph (FWD): from PH (placeholders) to loss tensor
        - function should return a dict with: PH, tensors, variables lists
            - dict keys should meet naming policy
                - list of special keys to consider while building fwd_func is under SPEC_KEYS
                - OPT part will be build only if there is a 'loss' tensor returned
                - dict may contain variables_lists and single variable under keys with 'var' in name
                    - variables returned under 'train_vars' key are optimized (if 'train_vars' key is not present all trainable vars are optimized)
                    - sub-lists of variables will serve for separate savers (saved in subfolders)
    opt_func:
        - rather should not be replaced, but if so then it should accept train_vars and gradients parameters
        - should return 'optimizer'

 NEModel class extends NNWrap with:
    - GPU automated management with multi-GPU training on towers (places model graph elements across available devices)
    - builds optimization (OPT) graph with default OPT function
        - calculates gradients for every tower >> averages them
        - AVT gradient clipping and scaled LR (warmup, annealing)
    - exemplary FWD graph function
"""

import numpy as np
import os
from typing import Optional, Callable, Tuple, List
import warnings
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

from pypaq.lipytools.little_methods import short_scin, stamp, get_params, get_func_dna
from pypaq.lipytools.moving_average import MovAvg
from pypaq.lipytools.pylogger import get_pylogger, get_hi_child
from pypaq.mpython.devices import mask_cuda, get_devices
from pypaq.pms.parasave import ParaSave
from pypaq.comoneural.nnwrap import NNWrap, NNWrapException
from pypaq.neuralmess.get_tf import tf
from pypaq.neuralmess.base_elements import num_var_floats, lr_scaler, gc_loss_reductor, log_vars, mrg_ckpts
from pypaq.neuralmess.layers import lay_dense
from pypaq.neuralmess.multi_saver import MultiSaver


# default FWD function (forward graph), it is given as an exemplary implementation
def fwd_graph(
        name: str,
        seq_len: int,
        emb_num: int=   50,
        emb_width: int= 100,
        n_labels: int=  2,
        seed: int=      321,
        baseLR=         0.003):
    with tf.variable_scope(name):

        in_PH = tf.placeholder(
            name=           'in_PH',
            dtype=          tf.int32,
            shape=          [None, seq_len])
        labels_PH = tf.placeholder(
            name=           'labels_PH',
            dtype=          tf.int32,
            shape=          [None, 1])
        emb = tf.get_variable( # some embeddings
            name=       'emb',
            shape=      [emb_num, emb_width],
            dtype=      tf.float32)

        feats = tf.nn.embedding_lookup(params=emb, ids=in_PH)

        logits = lay_dense(
            input=      feats,
            units=      n_labels,
            name=       'logits',
            seed=       seed)

        loss = tf.nn.sparse_softmax_cross_entropy_with_logits(
            labels= labels_PH,
            logits= logits)

        pred = tf.argmax(logits, axis=-1)

        acc = tf.reduce_mean(tf.cast(tf.equal(tf.cast(pred, dtype=tf.int32), labels_PH), dtype=tf.float32))

    return {
        'in_PH':    in_PH,
        'emb':      emb,
        'logits':   logits,
        'loss':     loss,
        'acc':      acc}

# default OPT function (optimization graph), should be used in most scenarios
def opt_graph(
        train_vars,
        gradients,
        opt_class=          tf.train.AdamOptimizer,
            # LR management (check pypaq.neuralmess.base_elements.lr_scaler)
        baseLR=             3e-4,
        warm_up=            None,
        ann_base=           None,
        ann_step=           1.0,
        n_wup_off: float=   2.0,
            # gradients clipping parameters (check pypaq.neuralmess.base_elements.gc_loss_reductor)
        clip_value=         None,
        avt_SVal=           0.1,
        avt_window=         100,
        avt_max_upd=        1.5,
        do_clip=            False,
        logger=             None):

    if not logger: logger = get_pylogger()

    g_step = tf.get_variable(  # global step variable
        name=           'g_step',
        shape=          [],
        trainable=      False,
        initializer=    tf.constant_initializer(0),
        dtype=          tf.int32)

    baseLR_var = tf.get_variable(  # base LR variable
        name=           'baseLR',
        shape=          [],
        trainable=      False,
        initializer=    tf.constant_initializer(baseLR),
        dtype=          tf.float32)

    scaled_LR = lr_scaler(
        baseLR=         baseLR_var,
        g_step=         g_step,
        warm_up=        warm_up,
        ann_base=       ann_base,
        ann_step=       ann_step,
        n_wup_off=      n_wup_off,
        logger=         logger)['scaled_LR']

    # updates with: optimizer, gg_norm, gg_avt_norm
    loss_reductorD = gc_loss_reductor(
        optimizer=      opt_class(learning_rate=scaled_LR),
        vars=           train_vars,
        g_step=         g_step,
        gradients=      gradients,
        clip_value=     clip_value,
        avt_SVal=       avt_SVal,
        avt_window=     avt_window,
        avt_max_upd=    avt_max_upd,
        do_clip=        do_clip,
        logger=         logger)

    # select OPT vars
    opt_vars = tf.global_variables(scope=tf.get_variable_scope().name)

    logger.debug(f' ### opt_vars: {len(opt_vars)} floats: {short_scin(num_var_floats(opt_vars))} ({opt_vars[0].device})')
    logger.log(5, log_vars(opt_vars))

    rd = {}
    rd.update({
        'g_step':       g_step,
        'baseLR_var':   baseLR_var,
        'scaled_LR':    scaled_LR,
        'opt_vars':     opt_vars})
    rd.update(loss_reductorD)
    return rd


class NEModelException(NNWrapException):
    pass

# NEModel Base class, implements most features (but not saving)
class NEModel(NNWrap):

    SPEC_KEYS = {
        'train_vars',   # list of variables to train (may be returned, otherwise all trainable are taken)
        'opt_vars',     # list of variables returned by opt_func
        'loss',         # loss
        'acc',          # accuracy
        'f1'}           # F1

    # some defaults of NEModel are in default opt_func (opt_graph) above
    INIT_DEFAULTS = {
        'seed':                 123,        # seed for TF and numpy
        'devices':              -1,         # : DevicesParam (check pypaq.mpython.devices)
            # training
        'batch_size':           64,         # training batch size
        'n_batches':            1000,       # default length of training
        'train_batch_IX':       0,          # default (starting) batch index (counter)
            # other
        'hpmser_mode':          False,      # it will set model to be read_only and quiet when running with hpmser
        'savers_names':         (None,),    # names of savers for MultiSaver
        'load_saver':           True,       # Optional[bool or str] for None/False does not load, for True loads default
        'read_only':            False,      # sets model to be read only - wont save anything (wont even create self.nnwrap_dir)
        'do_TB':                True,       # runs TensorBard, saves in self.nnwrap_dir
        'silent_TF_warnings':   False,      # turns off TF warnings
        'sep_device':           True,       # separate first device for variables, gradients_avg, optimizer (otherwise those ar placed on the first FWD calculations tower)
        'collocate_GWO':        False}      # collocates gradient calculations with tf.OPs (gradients are calculated on every tower with its operations, but remember that vars are on one device...) (otherwise with first FWD calculations tower)

    SAVE_FN_PFX = 'nemodel_dna' # filename (DNA) prefix

    def __init__(
            self,
            nngraph: Optional[Callable]=    None,       # forward function
            opt_func: Optional[Callable]=   opt_graph,  # optimization function building optimization (OPT) graph (from train_vars & gradients to optimizer)
            **kwargs):

        if kwargs.get('silent_TF_warnings', False):
            tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
            warnings.filterwarnings('ignore')

        self._dna_opt_func = {}
        self._gFWD = [] # list of dicts of all FWD graphs (from all devices)
        self._graph = None
        self._session = None
        self._saver = None

        NNWrap.__init__(
            self,
            nngraph=    nngraph,
            opt_func=   opt_func, # give with kwargs
            **kwargs)

    # ******************************************************************************************* NNWrap init submethods

    def _generate_name(
            self,
            given_name: Optional[str],
            timestamp: bool) -> str:
        name = f'NEModel_{self.nngraph.__name__}' if not given_name else given_name
        if timestamp: name += f'_{stamp()}'
        return name

    # overrides NNWrap version a little, since NEModel has TWO graph functions
    def _manage_dna(
            self,
            save_topdir: str,
            save_fn_pfx: str,
            **kwargs) -> None:

        # load dna from folder
        dna_saved = NEModel.load_dna(
            name=           self.name,
            save_topdir=    save_topdir,
            save_fn_pfx=    save_fn_pfx)

        # in case 'nngraph' was not given with init, try to get it from saved
        if not self.nngraph:
            self.nngraph = dna_saved.get('nngraph', None)

        if not self.nngraph:
            msg = 'nngraph was not given and has not been found in saved, cannot continue!'
            self._nwwlog.error(msg)
            raise NEModelException(msg)

        # get defaults of given nngraph and opt_func
        nngraph_func_params = get_params(self.nngraph)
        nngraph_func_params_defaults = nngraph_func_params['with_defaults']  # get nngraph defaults
        opt_func = kwargs['opt_func']
        opt_func_params_defaults = get_params(opt_func)['with_defaults'] if opt_func else {}
        if 'logger' in nngraph_func_params_defaults: nngraph_func_params_defaults.pop('logger')
        if 'logger' in opt_func_params_defaults: opt_func_params_defaults.pop('logger')

        self._dna.update(self.INIT_DEFAULTS)
        self._dna.update(opt_func_params_defaults)
        self._dna.update(nngraph_func_params_defaults)
        self._dna.update(dna_saved)
        self._dna.update(kwargs)  # update with kwargs (params of FWD & OPT) given NOW by user
        self._dna.update({
            'name':         self.name,
            'save_topdir':  save_topdir,
            'save_fn_pfx':  save_fn_pfx})

        dna_with_nwwlog = {}
        dna_with_nwwlog.update(self._dna)
        dna_with_nwwlog['logger'] = get_hi_child(
            logger= self._nwwlog,
            name=   f'{self.name}_sublogger')
        self._dna_nngraph = get_func_dna(self.nngraph, dna_with_nwwlog)
        self._dna_opt_func = get_func_dna(opt_func, dna_with_nwwlog)

        not_used_kwargs = {}
        for k in kwargs:
            if k not in self._dna_nngraph and k not in self._dna_opt_func:
                not_used_kwargs[k] = kwargs[k]

        self._nwwlog.debug(f'> {self.name} DNA sources:')
        self._nwwlog.debug(f'>> class INIT_DEFAULTS:        {self.INIT_DEFAULTS}')
        self._nwwlog.debug(f'>> nngraph defaults:           {nngraph_func_params_defaults}')
        self._nwwlog.debug(f'>> opt_func defaults:          {opt_func_params_defaults}')
        self._nwwlog.debug(f'>> DNA saved:                  {dna_saved}')
        self._nwwlog.debug(f'>> given kwargs:               {kwargs}')
        self._nwwlog.debug(f'> resolved DNA:')
        self._nwwlog.debug(f'nngraph complete DNA:          {self._dna_nngraph}')
        self._nwwlog.debug(f'opt_func complete DNA:         {self._dna_opt_func}')
        self._nwwlog.debug(f'>> kwargs not used by graphs : {not_used_kwargs}')
        self._nwwlog.debug(f'{self.name} complete DNA:      {self._dna}')

    def _manage_devices(self):

        self['devices'] = get_devices(
            devices=    self['devices'],
            namespace=  'TF1',
            logger=     get_hi_child(self._nwwlog, 'get_devices'))

        devices_other = []
        devices_gpu = []
        for device in self['devices']:
            if 'GPU' in device: devices_gpu.append(device)
            else: devices_other.append(device)

        # prepare ids & rewrite GPU devices
        ids = []
        if devices_gpu:
            devices_gpu = [f'/device:GPU:{ix}' for ix in range(len(devices_gpu))]
            ids = [dev[12:] for dev in devices_gpu]

        self['devices'] = devices_other + devices_gpu

        self._nwwlog.debug(f' > masking GPU devices: {ids}')
        mask_cuda(ids) # mask GPU devices

        # report devices
        if len(self['devices'])==1:
            if 'CPU' in self['devices'][0]: self._nwwlog.info(f'NEModel builds CPU device setup')
            else:                           self._nwwlog.info(f'NEModel builds single-GPU setup: {self["devices"]}')
        else:                               self._nwwlog.info(f'NEModel builds multi-dev setup for {len(self["devices"])} devices')

        if len(self['devices'])<3: self['sep_device'] = False # SEP is available for 3 or more devices

    # sets NNWrap seed in all possible areas
    def _set_seed(self) -> None:
        tf.set_random_seed(self['seed'])
        np.random.seed(self['seed'])

    # builds graph (FWD & OPT) and manages surroundings
    def _build_graph(self) -> None:

        # build FWD graph(s) >> manage variables >> build OPT graph
        self._graph = tf.Graph()
        with self._graph.as_default():

            self._nwwlog.debug(f'NEModel set TF & NP seed to {self["seed"]}')

            # builds graph @SEP, this graph wont be run, it is only needed to place variables, if not vars_sep >> variables will be placed with first tower
            if self['sep_device']:
                self._nwwlog.debug(f'NEModel places VARs on {self["devices"][0]}...')
                with tf.device(self['devices'][0]):
                    self.nngraph(**self._dna_nngraph)

            tower_devices = [] + self['devices']
            if self['sep_device']: tower_devices = tower_devices[1:] # trim SEP
            for dev in tower_devices:
                self._nwwlog.debug(f'NEModel builds FWD graph @device: {dev}')
                with tf.device(dev):
                    with tf.variable_scope('', reuse=tf.AUTO_REUSE):
                        self._gFWD.append(self.nngraph(**self._dna_nngraph))

            fwd_graph_return_dict = self._gFWD[0]
            self._nwwlog.debug(f'dictionary keys returned by fwd_func ({self.nngraph.__name__}): {fwd_graph_return_dict.keys()}')

            self.update(fwd_graph_return_dict) # update self with fwd_graph_return_dict

            # get FWD variables returned by fwd_func (4 saver)
            train_vars = [] # variables to train
            saver_vars = {} # dict of variables to save
            for key in self.get_all_fields():
                if 'var' in key.lower():
                    if key =='train_vars':
                        train_vars = self[key]
                        if type(train_vars) is not list: train_vars = [train_vars]
                    else:
                        if type(self[key]) is not list: saver_vars[key] = [self[key]]
                        else:                           saver_vars[key] = self[key]
            all_vars = tf.global_variables()

            # there are returned variables >> assert there are all variables returned in lists
            if saver_vars:
                all_vars_returned = []
                for key in saver_vars: all_vars_returned += saver_vars[key]
                there_are_all = True
                for var in all_vars:
                    if var not in all_vars_returned:
                        print(f' *** variable {var.name} not returned by fwd_func')
                        there_are_all = False
                assert there_are_all, 'ERR: there are some variables not returned by fwd_func in lists!'

            else: saver_vars['fwd_vars'] = all_vars # put all


            self._nwwlog.debug('NEModel variables to save from fwd_func:')
            for key in sorted(list(saver_vars.keys())):
                    varList = saver_vars[key]
                    if varList: self._nwwlog.log(5, f' ### vars @{key} - num: {len(varList)}, floats: {short_scin(num_var_floats(varList))} ({varList[0].device})')
                    else: self._nwwlog.log(5, ' ### no vars')
                    self._nwwlog.log(5, log_vars(varList))

            if 'loss' not in self: self._nwwlog.warning('NEModel: there is no loss in FWD graph, OPT graph wont be build!')
            if not self['opt_func']: self._nwwlog.warning(f'\nNEModel: OPT graph wont be build since opt_func is not given')

            # build optimization graph
            if self['opt_func'] and 'loss' in self:
                self._nwwlog.debug(f'Preparing OPT part with {self["opt_class"]}')

                # select trainable variables for OPT
                all_tvars = tf.trainable_variables()
                if train_vars:
                    # check if all train_vars are trainable:
                    for var in train_vars:
                        if var not in all_tvars:
                            self._nwwlog.warning(f'variable {var.name} is not trainable but is in train_vars, please check the graph!')
                else:
                    for key in saver_vars:
                        for var in saver_vars[key]:
                            if var in all_tvars:
                                train_vars.append(var)
                    assert train_vars, 'ERR: there are no trainable variables at the graph!'
                # log train_vars
                self._nwwlog.debug('NEModel trainable variables:')
                self._nwwlog.debug(f' ### train_vars: {len(train_vars)} floats: {short_scin(num_var_floats(train_vars))}')
                self._nwwlog.log(5, log_vars(train_vars))

                # build gradients for towers
                for ix in range(len(self._gFWD)):
                    tower = self._gFWD[ix]
                    tower['gradients'] = tf.gradients(
                        ys=                             tower['loss'],
                        xs=                             train_vars,
                        colocate_gradients_with_ops=    not self['collocate_GWO']) # TF default is False >> calculates gradients where OPS, for True >> where train_vars

                    # log gradients

                    # None_as_gradient case
                    device = 'UNKNOWN'
                    for t in tower['gradients']:
                        if t is not None:
                            device = t.device
                            break

                    self._nwwlog.debug(f' > gradients for {ix} tower got {len(tower["gradients"])} tensors ({device})')

                    self._nwwlog.log(5, 'NEModel variables and their gradients:')
                    for gix in range(len(tower['gradients'])):
                        grad = tower['gradients'][gix]
                        var = train_vars[gix]
                        self._nwwlog.log(5, f'{var} {var.device}')
                        self._nwwlog.log(5, f' > {grad}') # grad as a tensor displays device when printed (unless collocated with OP!)

                self['gradients'] = self._gFWD[0]['gradients']

                # check for None @ gradients
                none_grads = 0
                for grad in self['gradients']:
                    if grad is None: none_grads += 1
                if none_grads: self._nwwlog.warning(f'There are None gradients: {none_grads}/{len(self["gradients"])}, some trainVars may be unrelated to loss, please check the graph!')

                # average gradients
                if len(self['devices']) > 1:

                    self._nwwlog.debug(f'NEModel builds gradients averaging graph with device {self["devices"][0]} for {len(self._gFWD)} towers')
                    with tf.device(self["devices"][0]):
                        towerGrads = [tower['gradients'] for tower in self._gFWD]
                        avgGrads = []
                        for mGrads in zip(*towerGrads):
                            grads = []
                            for grad in mGrads:
                                if grad is not None: # None for variables not used while training now...
                                    expandedG = tf.expand_dims(input=grad, axis=-1)
                                    grads.append(expandedG)
                            if grads:
                                grad = tf.concat(values=grads, axis=-1)
                                grad = tf.reduce_mean(input_tensor=grad, axis=-1)
                                avgGrads.append(grad)
                            else: avgGrads.append(None)

                        self['gradients'] = avgGrads # update with averaged gradients
                        self._nwwlog.debug(f' > NEModel averaged gradients ({self["gradients"][0].device})')

                # finally build graph from elements
                with tf.variable_scope('OPT', reuse=tf.AUTO_REUSE):

                    self._nwwlog.debug(f'Building OPT graph for {self.name} model @device: {self["devices"][0]}')

                    with tf.device(self['devices'][0]):

                        opt_graph_return_dict = self['opt_func'](
                            train_vars=     train_vars,
                            gradients=      self['gradients'],
                            **self._dna_opt_func)
                        self._nwwlog.debug(f'dictionary keys returned by opt_func ({self["opt_func"].__name__}): {opt_graph_return_dict.keys()}')

                        self.update(opt_graph_return_dict)  # update self with opt_graph_return_dict

                        saver_vars['opt_vars'] = self['opt_vars']

        # create session
        config = tf.ConfigProto(allow_soft_placement=True)
        config.gpu_options.allow_growth = True
        self._session = tf.Session(
            graph=  self._graph,
            config= config)

        # create saver & load
        sv_keys = list(saver_vars.keys())
        for key in sv_keys:
            if not saver_vars[key]: saver_vars.pop(key) # remove keys with no variables (corner case, for proper saver)
        # add saver then load
        self._saver = MultiSaver(
            model_name= self.name,
            vars=       saver_vars,
            save_TFD=   self.save_topdir,
            savers=     self['savers_names'],
            session=    self._session,
            logger=     get_hi_child(self._nwwlog, 'MultiSaver'))
        if self['load_saver']: self.load_ckpt()

    def __call__(self, feed_dict:dict, fetch:List[str]) -> dict:
        out = self._session.run(feed_dict=feed_dict, fetches=[self[n] for n in fetch])
        return {n: e for n,e in zip(fetch,out)}

    def backward(self, feed_dict:dict, fetch:List[str]) -> dict:
        return self.__call__(feed_dict=feed_dict, fetch=fetch)

    # *********************************************************************************************** load / save / copy

    # reloads model checkpoint, updates baseLR
    def load_ckpt(self) -> None:
        saver = None if type(self['load_saver']) is bool else self['load_saver']
        self._saver.load(saver=saver)
        if 'baseLR' in self: self.update_baseLR(self['baseLR'])

    # saves model checkpoint
    def save_ckpt(self) -> None:
        assert not self['read_only'], f'ERR: cannot save NEModel checkpoint {self.name} while model is readonly!'
        self._saver.save()

    @classmethod
    def copy_checkpoint(
            cls,
            name_src: str,
            name_trg: str,
            save_topdir_src: Optional[str]= None,
            save_topdir_trg: Optional[str]= None):

        if not save_topdir_src: save_topdir_src = cls.SAVE_TOPDIR
        if not save_topdir_trg: save_topdir_trg = save_topdir_src

        nm_SFD = f'{save_topdir_src}/{name_src}'
        ckptL = [cfd for cfd in os.listdir(nm_SFD) if os.path.isdir(os.path.join(nm_SFD, cfd))]
        if 'opt_vars' in ckptL: ckptL.remove('opt_vars')
        for ckpt in ckptL:
            mrg_ckpts(
                ckptA=          ckpt,
                ckptA_FD=       nm_SFD,
                ckptB=          None,
                ckptB_FD=       None,
                ckptM=          ckpt,
                ckptM_FD=       f'{save_topdir_trg}/{name_trg}',
                replace_scope=  name_trg)

    # *************************************************************************************************************** GX

    @classmethod
    def gx_ckpt(
            cls,
            name_A: str,                        # name parent A
            name_B: str,                        # name parent B
            name_child: str,                    # name child
            save_topdir_A: Optional[str]=       None,
            save_topdir_B: Optional[str]=       None,
            save_topdir_child: Optional[str]=   None,
            ratio: float=                       0.5,
            noise: float=                       0.03):

        if not save_topdir_A: save_topdir_A = cls.SAVE_TOPDIR
        if not save_topdir_B: save_topdir_B = save_topdir_A
        if not save_topdir_child: save_topdir_child = save_topdir_A

        mfd = f'{save_topdir_A}/{name_A}'
        ckptL = [dI for dI in os.listdir(mfd) if os.path.isdir(os.path.join(mfd,dI))]
        if 'opt_vars' in ckptL: ckptL.remove('opt_vars')

        for ckpt in ckptL:
            mrg_ckpts(
                ckptA=          ckpt,
                ckptA_FD=       f'{save_topdir_A}/{name_A}/',
                ckptB=          ckpt,
                ckptB_FD=       f'{save_topdir_B}/{name_B}/',
                ckptM=          ckpt,
                ckptM_FD=       f'{save_topdir_child}/{name_child}/',
                replace_scope=  name_child,
                ratio=          ratio,
                noise=          noise)

    # ***************************************************************************************************** train / test

    # builds feed dict from given batch of data
    def build_feed(self, batch: dict, train=True) -> dict:
        self._nwwlog.warning('NEModel.build_feed() should be overridden!')
        return {}

    # TODO: refactor according to MOTorch train concept
    #  - load_data()
    def run_train(
            self,
            n_batches: Optional[int]=   None,
            test_freq=                  100,
            mov_avg_factor=             0.1,
            save_max=                   True,
            use_F1=                     True, # TODO: use
            **kwargs) -> float:

        if not self._batcher: raise NEModelException('NEModel has not been given data for training, use load_data() or give it while training!')

        self._nwwlog.info(f'{self.name} - training starts [acc / F1 / loss]')
        self._nwwlog.info(f'data sizes (TR,VL,TS) samples: {self._batcher.get_data_size()}')

        if n_batches is None: n_batches = self['n_batches']  # take default
        self._nwwlog.info(f'batch size:             {self["batch_size"]}')
        self._nwwlog.info(f'train for num_batches:  {n_batches}')

        batch_IX = 0
        tr_accL = []
        tr_f1L = []
        tr_lssL = []

        score_name = 'F1' if use_F1 else 'acc'
        ts_score_max = 0  # test score (acc or F1) max
        ts_score_all_results = []  # test score all results
        ts_score_mav = MovAvg(mov_avg_factor)  # test score (acc or F1) moving average

        ts_bIX = [bIX for bIX in range(n_batches+1) if not bIX % test_freq] # batch indexes when test will be performed
        assert ts_bIX, 'ERR: model SHOULD BE tested while training!'
        ten_factor = int(0.1*len(ts_bIX)) # number of tests for last 10% of training
        if ten_factor < 1: ten_factor = 1 # we need at least one result
        if self['hpmser_mode']: ts_bIX = ts_bIX[-ten_factor:]

        while batch_IX < n_batches:

            batch = self._batcher.get_batch()
            feed = self.build_feed(batch)
            out = self(feed_dict=feed, fetch=['optimizer','loss','acc','f1','gg_norm','gg_avt_norm'])
            out.pop('optimizer')

            batch_IX += 1
            self['train_batch_IX'] += 1

            if self['do_TB']:
                for k in out:
                    self.log_TB(value=out[k], tag=f'tr/{k}', step=self['train_batch_IX'])
            tr_accL.append(out['acc'])
            tr_f1L.append(out['f1'])
            tr_lssL.append(out['loss'])

            if batch_IX in ts_bIX:
                ts_acc, ts_f1, ts_loss = self.run_test()

                ts_score = ts_f1 if use_F1 else ts_acc
                if ts_score is not None:
                    ts_score_all_results.append(ts_score)
                if self['do_TB']:
                    self.log_TB(value=ts_loss,                      tag='ts/loss',              step=self['train_batch_IX'])
                    self.log_TB(value=ts_acc,                       tag='ts/acc',               step=self['train_batch_IX'])
                    self.log_TB(value=ts_f1,                        tag='ts/F1',                step=self['train_batch_IX'])
                    self.log_TB(value=ts_score_mav.upd(ts_score),   tag=f'ts/{score_name}_mav', step=self['train_batch_IX'])

                tr_acc_nfo = f'{100*sum(tr_accL)/test_freq:.1f}'
                tr_f1_nfo =  f'{100*sum(tr_f1L)/test_freq:.1f}'
                ts_acc_nfo = f'{100*ts_acc:.1f}'
                ts_f1_nfo = f'{100*ts_f1:.1f}'
                self._nwwlog.info(f'# {self["train_batch_IX"]:5d} TR: {tr_acc_nfo} / {tr_f1_nfo} / {sum(tr_lssL)/test_freq:.3f} -- TS: {ts_acc_nfo} / {ts_f1_nfo} / {ts_loss:.3f}')

                tr_accL = []
                tr_f1L = []
                tr_lssL = []

                if ts_score is not None and ts_score > ts_score_max:
                    ts_score_max = ts_score
                    if not self['read_only'] and save_max: self.save_ckpt() # model is saved for max ts_score

        # weighted (linear ascending weight) test score for last 10% test results
        ts_score_wval = None
        if ts_score_all_results:
            ts_score_wval = 0.0
            weight = 1
            sum_weight = 0
            for tr in ts_score_all_results[-ten_factor:]:
                ts_score_wval += tr*weight
                sum_weight += weight
                weight += 1
            ts_score_wval /= sum_weight

            if self['do_TB']: self.log_TB(value=ts_score_wval, tag=f'ts/ts_{score_name}_wval', step=self['train_batch_IX'])

        self._nwwlog.info(f'### model {self.name} finished training')
        if ts_score_wval is not None:
            self._nwwlog.info(f' > test_{score_name}_max:  {ts_score_max:.4f}')
            self._nwwlog.info(f' > test_{score_name}_wval: {ts_score_wval:.4f}')

        return ts_score_wval

    def run_test(self, **kwargs) -> Tuple[Optional[float], Optional[float], float]:

        if not self._batcher: raise NEModelException('NEModel has not been given data for testing, use load_data() or give it while testing!')

        batches = self._batcher.get_TS_batches()
        lossL = []
        accL = []
        f1L = []
        for batch in batches:
            feed = self.build_feed(batch)
            out = self(feed_dict=feed, fetch=['loss','acc','f1'])
            lossL.append(out['loss'])
            accL.append(out['acc'])
            f1L.append(out['f1'])

        return sum(accL)/len(accL), sum(f1L)/len(f1L), sum(lossL)/len(lossL)

    @property
    def gFWD(self):
        return self._gFWD

    @property
    def session(self):
        return self._session

    # updates baseLR in graph - but not saves it to the checkpoint
    def update_baseLR(self, lr: Optional[float]) -> None:
        if 'baseLR_var' not in self:
            self._nwwlog.warning('NEModel: There is no LR variable in graph to update')
        else:
            if lr is not None:
                old = self['baseLR']
                self['baseLR'] = lr
                self._nwwlog.debug(f'NEModel {self.name} updated baseLR from {old} to {self["baseLR"]}')
            self._session.run(tf.assign(ref=self['baseLR_var'], value=self['baseLR']))
            self._nwwlog.debug(f'NEModel {self.name} updated baseLR_var (graph variable) with baseLR: {self["baseLR"]}')

    def __str__(self):
        # TODO: add some NEModel-specific output (name, graph info, weights)
        return ParaSave.__str__(self)


from typing import List

from pypaq.neuralmess.base_elements import tf
from pypaq.neuralmess.layers import lay_dense, zeroes


def a2c_graph(
        name=               'a2c',
        observation_width=  4,
        num_actions=        2,
        two_towers=         False,  # builds separate towers for Actor & Critic
        num_layers=         1,
        layer_width=        50,
        lay_norm=           False,
        use_scaled_ce=      True,   # for True uses experimental scaled ce loss for Actor
        use_huber=          False,  # for True uses Huber loss for Critic
        seed=               123,
        verb=               0):

    with tf.variable_scope(name):

        observation_PH = tf.placeholder(
            shape=  (None, observation_width),
            dtype=  tf.float32,
            name=   'observation')

        action_PH = tf.placeholder(
            shape=  None,
            dtype=  tf.int32,
            name=   'action')

        return_PH = tf.placeholder(
            shape=  None,
            dtype=  tf.float32,
            name=   'return')

        zsL: List[tf.Tensor] = []  # list of zeroes Tensors

        hidden_layers = tuple([layer_width] * num_layers)

        inp = tf.keras.layers.LayerNormalization(axis=-1)(observation_PH) if lay_norm else observation_PH

        layer = inp
        for i in range(len(hidden_layers)):
            layer = lay_dense(
                input=      layer,
                name=       f'hidden_layer_{i}',
                units=      hidden_layers[i],
                activation= tf.nn.relu,
                seed=       seed)
            zsL.append(zeroes(layer))
            if lay_norm: layer = tf.keras.layers.LayerNormalization(axis=-1)(layer)

        layer_second_tower = layer
        if two_towers:
            layer_second_tower = inp
            for i in range(len(hidden_layers)): # INFO: shape of second tower will be as first one
                layer_second_tower = lay_dense(
                    input=      layer_second_tower,
                    name=       f'hidden_layer_st_{i}',
                    units=      hidden_layers[i],
                    activation= tf.nn.relu,
                    seed=       seed)
                zsL.append(zeroes(layer_second_tower))
                if lay_norm: layer_second_tower = tf.keras.layers.LayerNormalization(axis=-1)(layer_second_tower)

        # Value from critic
        value = lay_dense(
            input=      layer_second_tower,
            name=       'value',
            units=      1,
            activation= None,
            seed=       seed)
        if verb > 0: print(f' > value: {value}')

        advantage = return_PH - value
        if verb > 0: print(f' > advantage: {advantage}')

        # action logits of Actor policy
        action_logits = lay_dense(
            input=      layer,
            name=       'action_logits',
            units=      num_actions,
            activation= None,
            seed=       seed)
        if verb > 0: print(f' > action_logits: {action_logits}')
        probs = tf.nn.softmax(action_logits)

        # ************************************************************************************************** loss definition

        if use_scaled_ce:

            action_prob_selected = tf.gather(params=probs, indices=action_PH, axis=1, batch_dims=1)
            #action_prob_selected = tf.squeeze(action_prob_selected)
            if verb>0: print(f' > action_prob_selected: {action_prob_selected}')

            # crossentropy loss
            actor_ce = -tf.math.log(action_prob_selected)
            actor_ce_neg = -tf.math.log(1-action_prob_selected)

            # select loss for positive and negative advantage
            actor_ce = tf.where(
                condition=  tf.greater(advantage,0),
                x=          actor_ce,
                y=          actor_ce_neg)

            loss_actor_weighted_mean = tf.reduce_mean(actor_ce * tf.math.abs(advantage))  # scale loss with advantage

        else:

            actor_ce = tf.nn.sparse_softmax_cross_entropy_with_logits(logits=action_logits, labels=action_PH)
            loss_actor_weighted_mean = tf.reduce_mean(actor_ce * advantage)

        if verb > 0: print(f' > actor_ce: {actor_ce}')
        actor_ce_mean = tf.reduce_mean(actor_ce)

        if use_huber:
            huber_loss = tf.keras.losses.Huber()
            loss_critic = huber_loss(value, return_PH)
        else:
            loss_critic = advantage * advantage # MSE

        if verb > 0: print(f' > loss_critic: {loss_critic}')
        loss_critic_mean = tf.reduce_mean(loss_critic)

        loss = loss_actor_weighted_mean + loss_critic_mean
        if verb > 0: print(f' > loss: {loss}')

    return {
        'observation_PH':   observation_PH,
        'action_PH':        action_PH,
        'return_PH':        return_PH,

        'value':            value,
        'advantage':        advantage,
        'action_logits':    action_logits,
        'probs':            probs,

        'actor_ce':         actor_ce,
        'actor_ce_mean':    actor_ce_mean,

        'loss':             loss,
        'loss_actor':       loss_actor_weighted_mean,
        'loss_critic':      loss_critic_mean,

        'zeroes':           zsL}


"""

 2019 (c) piteren

    some encoders

"""

from pypaq.neuralmess.get_tf import tf
from pypaq.neuralmess.base_elements import my_initializer, list_of_layers
from pypaq.neuralmess.layers import zeroes, lay_dense, lay_conv1D, attn, tf_drop

# TODO: probably in tf 1.15.2 histogram 'family=' should be replaced with 'description=' << check it on histogram case


# DRT layer (based on Transformer scaling two denses concept)
def lay_DRT(
        input,
        name=           'lay_DRT',  # scope name, be careful when stacked since auto_reuse
        hist_name=      None,       # family name of histogram
        do_scaled_dns=  False,      # two denses (True) or single dense (False)
        dns_scale=      4,          # scale for two denses
        activation=     tf.nn.relu, # gelu is really worth a try
        lay_dropout=    0.0,        # dropout after dense/s
        add_res=        True,       # residual yes/no
        res_dropout=    0.0,        # dropout on residual connection
        training_flag=  None,       # training flag tensor (for dropout)
        initializer=    None,
        seed=           12321):

    if not hist_name: hist_name = name
    lay_width = input.shape[-1]
    if initializer is None: initializer = my_initializer(seed)
    hist_summ = []

    with tf.variable_scope(name_or_scope=name, reuse=tf.AUTO_REUSE):

        # layer_norm
        output = tf.keras.layers.LayerNormalization(axis=-1)(input)

        hist_summ.append(tf.summary.histogram('a_input', output, family=hist_name))

        # two denses (up-scale >> down-scale - transformer like)
        if do_scaled_dns:

            # dense (scale up), no activation
            output = lay_dense(
                input=          output,
                units=          int(lay_width * dns_scale),
                activation=     None,
                use_bias=       True,
                initializer=    initializer,
                seed=           seed,
                name=           'denseS')
            hist_summ.append(tf.summary.histogram('b_denseUp', output, family=hist_name))

            # activation
            output = activation(output)
            zsL = [zeroes(output)]  # zeroes list
            hist_summ.append(tf.summary.histogram('c_activation', output, family=hist_name))

            # dense (scale down), no activation
            output = lay_dense(
                input=          output,
                units=          lay_width,
                activation=     None,
                name=           'DRTdenseNA',
                use_bias=       True,
                initializer=    initializer,
                seed=           seed)
            hist_summ.append(tf.summary.histogram('d_denseDown', output, family=hist_name))

        # just dense
        else:
            output = lay_dense(
                input=          output,
                units=          lay_width,
                activation=     activation,
                use_bias=       True,
                initializer=    initializer,
                seed=           seed,
                name=           'dense')
            zsL = [zeroes(output)]  # zeroes list

        # layer dropout
        if lay_dropout:
            output = tf.layers.dropout(
                inputs=     output,
                rate=       lay_dropout,
                training=   training_flag,
                seed=       seed)

        if add_res:
            # residual dropout
            if res_dropout:
                input = tf.layers.dropout(
                    inputs=     input,
                    rate=       res_dropout,
                    training=   training_flag,
                    seed=       seed)
            output += input # residual
            hist_summ.append(tf.summary.histogram('e_residual', output, family=hist_name))

    return{
        'output':       output,
        'hist_summ':    hist_summ,
        'zeroes':       zsL}

# Deep Residual encoder based on stacked lay_DRT
def enc_DRT(
        input,
        name=               'enc_DRT',
        in_dropout=         0.0,        # dropout on input
        shared_lays :bool=  False,      # shared variables in enc_layers
        n_layers=           6,
        lay_width :int=     None,       # for None matches input width
        do_scaled_dns=      True,
        dns_scale=          4,          # scale(*) of first dense
        activation=         tf.nn.relu, # gelu is really worth a try
        lay_dropout=        0.0,        # dropout after two denses
        add_res=            True,
        res_dropout=        0.0,        # dropout on residual
        training_flag=      None,       # training flag tensor (for dropout)
        initializer=        None,
        seed=               12321,
        n_hist=             4,          # number of histogram layers (for TB)
        verb=               0):

    input_width = input.shape[-1]
    lay_width_matched = ''
    if lay_width is None:
        lay_width = input_width
        lay_width_matched = '(lay_width taken form input width)'
    if verb>0: print(f'\nBuilding DRTencoder ({n_layers}x{lay_width}) {lay_width_matched}...')

    if initializer is None: initializer = my_initializer(seed)

    hist_summ = []
    hist_layers = list_of_layers(n_layers, n_select=n_hist)
    if verb>1: print(' > histogram layers of DRTencoder:', hist_layers)

    zsL = [] # zeroes list
    with tf.variable_scope(name):

        # input dropout
        if in_dropout:
            input = tf.layers.dropout(
                inputs=     input,
                rate=       res_dropout,
                training=   training_flag,
                seed=       seed)

        # input projection, no activation
        if input_width != lay_width:
            input = lay_dense(
                input=          input,
                units=          lay_width,
                use_bias=       False,
                initializer=    initializer,
                seed=           seed)
            if verb>0: print(f'projected input to layWidth({lay_width}) since it differs({input_width})')

        input = tf.keras.layers.LayerNormalization(axis=-1)(input) # input layer_norm

        output = input # for 0 layers case
        for nL in range(n_layers):

            lay_name = f'DRLay_{nL}' if not shared_lays else 'DRLay_shared'
            lay_out = lay_DRT(
                input=          output,
                name=           lay_name,
                hist_name=      name,
                do_scaled_dns=  do_scaled_dns,
                dns_scale=      dns_scale,
                activation=     activation,
                lay_dropout=    lay_dropout,
                add_res=        add_res,
                res_dropout=    res_dropout,
                training_flag=  training_flag,
                initializer=    initializer,
                seed=           seed)

            output = lay_out['output']
            if nL in hist_layers: hist_summ.append(lay_out['hist_summ'])
            zsL += lay_out['zeroes']

    return {
        'output':       output,
        'hist_summ':    hist_summ,
        'zeroes':       zsL}

# CNN 1D encoder (for sequences, LN > CNN > act > drop > RES), number of parameters: n_layers*kernel*input.width*n_filters
def enc_CNN(
        input :tf.Tensor,
        history :tf.Tensor=                 None,       # optional history(state) tensor with shape [bsz, n_layers ,kernel-1, n_filters], >> masked cnn
        name=                               'enc_CNN',
        time_drop=                          0.0,
        feat_drop=                          0.0,
        # layer
        shared_lays :bool=                  False,      # shared variables in enc_layers
        n_layers :int=                      12,         # num of layers
        kernel :int=                        3,          # layer kernel
        n_filters :int=                     128,        # num of filters
        activation=                         tf.nn.relu, # global enc activation func, gelu is really worth a try
        lay_drop : float or None=           0.0,
        # lay_DRT
        do_ldrt=                            False,      # lay DRT - build or not
        ldrt_scaled_dns=                    True,       # lay DRT - build scaled two denses
        ldrt_scale: int or None=            4,          # lay DRT scale of first dense
        ldrt_drop: float or None=           0.0,        # lay DRT dropout
        ldrt_res=                           True,       # lay DRT - do residual connection
        ldrt_res_drop :float or None=       0.0,        # lay DRT residual dropout

        training_flag :tf.Tensor or bool=   None,       # dropout training flag tensor
        initializer=                        None,
        seed :int=                          12321,
        n_hist :int=                        4,          # number of histogram layers
        verb=                               0):

    if verb>0: print(f'\n *** enc_CNN *** Building {name} ({n_layers}x{n_filters})...')

    if initializer is None: initializer = my_initializer(seed)

    # manage history
    history_lays = None
    if history is not None:
        history_lays = tf.unstack(history, axis=-3)
        if verb>1: print(f' > state_lays len {len(history_lays)} of: {history_lays[0]}')

    hist_summ = []
    hist_layers = list_of_layers(n_layers, n_select=n_hist)
    if verb>1: print(f' > histogram layers of cnn encoder: {hist_layers}')

    with tf.variable_scope(name, reuse=tf.AUTO_REUSE):

        input_lays = [] # here we will store inputs of the following layers to extract the state (history)
        zsL = [] # zeroes

        # tf dropout
        if time_drop or feat_drop:
            input = tf_drop(
                input=      input,
                time_drop=  time_drop,
                feat_drop=  feat_drop,
                train_flag= training_flag,
                seed=       seed)

        # input projection - to match n_filters and input width
        if verb>1: print(f' > encoder input: {input}')
        if input.shape[-1] != n_filters:
            input = lay_dense(
                input=          input,
                units=          n_filters,
                name=           'enc_input_projection',
                initializer=    initializer)
            if verb>1: print(f' > encoder projected input: {input}')

        output = input # for 0 layers case
        sub_output = input # first input
        for depth in range(n_layers):

            lay_name = f'enc_CNN_lay_{depth}' if not shared_lays else 'enc_CNN_lay_shared'
            if verb>1: print(f'<< layer {lay_name}:')

            lay_input = tf.concat([history_lays[depth], sub_output], axis=-2) if history_lays else sub_output
            if verb>1:
                print(f' > sub_output (previous): {sub_output}')
                print(f' > lay_input (eventually padded): {lay_input}')
            input_lays.append(lay_input)

            hist_lay = depth in hist_layers

            with tf.variable_scope(lay_name):

                if hist_lay: hist_summ.append(tf.summary.histogram('a_lay_in', lay_input, family=name))

                # LN
                lay_input = tf.keras.layers.LayerNormalization(axis=-1)(lay_input)
                if hist_lay: hist_summ.append(tf.summary.histogram('b_LN', lay_input, family=name))

                # conv no activation
                output = lay_conv1D(
                    input=          lay_input,
                    name=           'conv1D',
                    kernels=        kernel,
                    filters=        n_filters,
                    activation=     None,
                    initializer=    initializer,
                    padding=        'same' if history is None else 'valid',
                    seed=           seed,
                    verb=           0)
                if hist_lay: hist_summ.append(tf.summary.histogram('c_cnn', output, family=name))

                # activation
                if activation:
                    output = activation(output)
                    zsL += [zeroes(output)]  # catch zeroes
                    if hist_lay: hist_summ.append(tf.summary.histogram('d_activation', output, family=name))

                # dropout
                if lay_drop:
                    output = tf.layers.dropout(
                        inputs=     output,
                        rate=       lay_drop,
                        training=   training_flag,
                        seed=       seed)
                    if hist_lay: hist_summ.append(tf.summary.histogram('e_drop', output, family=name))

                # RES, here we take sub_output, since lay_input may be padded by history
                output += sub_output
                if hist_lay: hist_summ.append(tf.summary.histogram('f_residual', output, family=name))

                if verb>1: print(f' > output (layer): {output}')

                if do_ldrt:
                    lay_out = lay_DRT(
                        input=          output,
                        name=           lay_name + '_lay_DRT',
                        hist_name=      name,
                        do_scaled_dns=  ldrt_scaled_dns,
                        dns_scale=      ldrt_scale,
                        activation=     activation,
                        lay_dropout=    ldrt_drop,
                        add_res=        ldrt_res,
                        res_dropout=    ldrt_res_drop,
                        training_flag=  training_flag,
                        initializer=    initializer,
                        seed=           seed)
                    output = lay_out['output']
                    zsL += lay_out['zeroes']
                    if hist_lay: hist_summ.append(lay_out['hist_summ'])

                sub_output = output

    output = tf.keras.layers.LayerNormalization(axis=-1)(output) # final LN

    # prepare fin_state
    fin_state = None
    if history is not None:
        state = tf.stack(input_lays, axis=-3)
        if verb>1: print(f' > state (stacked): {state}')
        fin_state = tf.split(state, num_or_size_splits=[-1, kernel-1], axis=-2)[1]
        if verb>1: print(f' > fin_state (split): {fin_state}')

    if verb>1: print(f' > {name} output: {output}')
    return {
        'output':       output,
        'state':        fin_state, # history for next
        'hist_summ':    hist_summ,
        'zeroes':       zsL}

# TODO: is family_name in tblock correctly used (following tblocks...)?
# transformer encoder
def enc_TNS(
        in_seq,                                         # input sequence embeddings [batch, seq, emb]
        name=                               'enc_TNS',
        seq_out :bool=                      True,       # transformer seq2seq, if False seq2one (Task Attention Transformer)
        add_PE :bool=                       True,       # add positional embeddings
        do_LN :bool=                        True,       # do layer norm in blocks
        shared_lays :bool=                  False,      # shared variables in blocks
        n_blocks=                           12,
        n_heads=                            8,
        dense_mul :int or float=            4,          # dense (after att) scale
        activation=                         tf.nn.relu, # dense activation
        max_seq_len=                        100,        # used only to set shape (axis 0) of positional embeddings
        dropout=                            0.0,        # dropout of FC after attention
        dropout_att=                        0.0,        # dropout of attention probabilities
        training_flag :tf.Tensor or bool=   None,       # training flag (bool or tensor) for dropout
        initializer=                        None,
        seed=                               12321,
        n_hist=                             4,          # number of histogram layers
        verb=                               0):

    if initializer is None: initializer = tf.truncated_normal_initializer(stddev=0.01, seed=seed)

    # split feats(-1) for heads
    def split_heads(x, n_heads):
        x = tf.split(x, n_heads, axis=-1) # list of tensors
        return tf.stack(x, axis=-3)  # [batch, head, seq, feats]

    # merge heads over feats(-1)
    def merge_heads(x):
        x = tf.unstack(x, axis=-3)
        return tf.concat(x, axis=-1)

    # multi_head_attention for input
    def mh_attn(
            in_seq,         # input sequence [batch, seq, feats]
            query,          # None for self attention, otherwise TAT [batch,1,feats]
            n_heads,
            dropout_att,
            training_flag,
            initializer,
            seed):

        # input projection of in_seq for KQV or KV(if query)
        width = in_seq.shape[-1].value
        proj_size = 3 if query is None else 2
        c = lay_dense(
            input=          in_seq,  # [batch, seq, feats]
            units=          width*proj_size,
            name=           'mhProj',
            activation=     None,
            initializer=    initializer,
            seed=           seed)
        ins_split = tf.split(c, proj_size, axis=-1) # split projected

        if query is not None:
            #q = query # projection for Q is not needed (at least with 1 head) # TODO: here previous version without projection
            q = lay_dense(
                input=          query,  # [batch, seq, feats]
                units=          width,
                name=           'mhProj_query',
                activation=     None,
                initializer=    initializer,
                seed=           seed)
            k,v = ins_split
        else:
            q,k,v = ins_split
        q = split_heads(q, n_heads)
        k = split_heads(k, n_heads)
        v = split_heads(v, n_heads)

        # attention
        att_out = attn(
            q=          q,
            k=          k,
            v=          v,
            dropout=    dropout_att,
            drop_flag=  training_flag,
            seed=       seed)
        a = att_out['attention']
        a = merge_heads(a)
        return {
            'attention':    a,
            'att_vals':     att_out['att_weights']}

    # transformer block
    def tblock(
            in_seq,
            task_query,
            do_LN,
            n_heads,
            dropout_att,
            dense_mul,
            dense_activation,
            dropout,
            training_flag,
            initializer,
            family_name,
            seed):

        hist_summ = []

        output = in_seq
        if task_query is None:
            hist_summ.append(tf.summary.histogram('a_inputSeq', output, family=family_name))
            # layer norm on seq
            if do_LN:
                output = tf.keras.layers.LayerNormalization(axis=-1)(output)
                hist_summ.append(tf.summary.histogram('b_inputSeqLN', output, family=family_name))
        else:
            hist_summ.append(tf.summary.histogram('a_inTaskQuery', task_query, family=family_name))
            # layer norm on task_query
            if do_LN:
                task_query = tf.keras.layers.LayerNormalization(axis=-1)(task_query)
                hist_summ.append(tf.summary.histogram('b_taskQueryLN', task_query, family=family_name))

        # multi-head self attention
        mha_out = mh_attn(
            in_seq=         output,
            query=          task_query,
            n_heads=        n_heads,
            dropout_att=    dropout_att,
            training_flag=  training_flag,
            initializer=    initializer,
            seed=           seed)
        output =    mha_out['attention']
        att_vals =  mha_out['att_vals']
        hist_summ.append(tf.summary.histogram('c_mhAttn', output, family=family_name))

        # dense without activation
        output = lay_dense(
            input=          output,
            units=          output.shape[-1].value,
            name=           'afterAttProj',
            initializer=    initializer,
            seed=           seed)
        hist_summ.append(tf.summary.histogram('d_denseAftAtt', output, family=family_name))

        if dropout:
            output = tf.layers.dropout(
                inputs=     output,
                rate=       dropout,
                training=   training_flag,
                seed=       seed)

        # residual 1
        if task_query is None:
            res1_out = in_seq + output
            hist_summ.append(tf.summary.histogram('e_res_onInputSeq', res1_out, family=family_name))
        else:
            res1_out = task_query + output
            hist_summ.append(tf.summary.histogram('e_res_onTaskQuery', res1_out, family=family_name))

        out_drtlay = lay_DRT(
            input=          res1_out,
            do_scaled_dns=  True,
            dns_scale=      dense_mul,
            activation=     dense_activation,
            lay_dropout=    dropout,
            res_dropout=    0.0, # TODO: make experiments
            training_flag=  training_flag,
            initializer=    initializer,
            seed=           seed)
        output =        out_drtlay['output']
        hist_summ +=    out_drtlay['hist_summ']
        zsL =           out_drtlay['zeroes']

        return {
            'output':       output,
            'hist_summ':    hist_summ,
            'att_vals':     att_vals,
            'zeroes':       zsL}

    width = in_seq.shape[-1]        # sequence width (feats)
    seq_len = tf.shape(in_seq)[-2]  # sequence length (time)

    if verb>0:
        print(f'\nBuilding {name} (transformer encoder) ({n_blocks}x{width}, denseMul {dense_mul:.1f}), ')
        print(f' > dropout: {dropout:.2f} {dropout_att:.2f}(att)')
        print(' > seq2seq mode...') if seq_out else print(' > task attention mode...')

    hist_layers = list_of_layers(n_blocks, n_select=n_hist)
    if verb>1: print(' > histogram layers of transformer encoder:', hist_layers)

    with tf.variable_scope(name):

        hist_summ = []  # list of histogram summaries

        if verb>1: print(' > transformer input', in_seq)
        hist_summ.append(tf.summary.histogram('a_transformerInput', in_seq, family=name))

        # init task_query (for first block - input averaged over time (seq))
        task_query = None
        if not seq_out:
            if do_LN: in_seq = tf.keras.layers.LayerNormalization(axis=-1)(in_seq)
            task_query = tf.reduce_mean(in_seq, axis=-2, keep_dims=True)# [batch,1,feats]
            """ experimental variable for task_query
            task_query = tf.get_variable(
                name=           'task_query',
                shape=          [1,width],
                initializer=    initializer)
            """
            if verb>1: print(' > first task_query (reduced input) for TAT', task_query)

        # positional embedding
        if add_PE:
            pos_emb_var = tf.get_variable(
                name=           'tnsPosEmb',
                shape=          [max_seq_len, width],
                initializer=    initializer)
            in_seq += tf.nn.embedding_lookup(params=pos_emb_var, ids=tf.range(seq_len))
            if verb>1: print(' > added positional embedding to the input...')
            hist_summ.append(tf.summary.histogram('b_transformerPosEmbInput', in_seq, family=name))

        if verb>1: print(f' > building {n_blocks} blocks of transformer...')
        att_vals = [] # list of block attention values
        zsL = []
        block_output = None
        for nB in range(n_blocks):
            hist_lay = nB in hist_layers
            lay_name = f'block_{nB}' if not shared_lays else 'block_shared'
            with tf.variable_scope(lay_name, reuse=tf.AUTO_REUSE):
                bo_dict = tblock(
                    in_seq=             in_seq,
                    task_query=         task_query,
                    do_LN=              do_LN,
                    n_heads=            n_heads,
                    dropout_att=        dropout_att,
                    dense_mul=          dense_mul,
                    dense_activation=   activation,
                    dropout=            dropout,
                    training_flag=      training_flag,
                    initializer=        initializer,
                    family_name=        name,
                    seed=               seed)
                block_output = bo_dict['output']
                if task_query is None: in_seq = block_output
                else: task_query = block_output

                zsL +=                      bo_dict['zeroes']
                if hist_lay: hist_summ +=   bo_dict['hist_summ']
                att_block_vals =            bo_dict['att_vals'] #[batch,head,query_n or seq,seq]
                att_vals.append(att_block_vals)

        if task_query is None:  output = block_output
        else:                   output = tf.squeeze(task_query, axis=-2)

        if do_LN: output = tf.keras.layers.LayerNormalization(axis=-1)(output)

        hist_summ.append(tf.summary.histogram('c_transformer_out', output, family=name))

    if verb>1: print(f' > {name} output {output}')
    return {
        'output':       output,
        'hist_summ':    hist_summ,
        'att_vals':     att_vals,
        'zeroes':       zsL}


if __name__ == "__main__":

    n_lays = 3
    width = 50
    kernel = 5

    input_PH = tf.placeholder(
        name=   'input_PH',
        dtype=  tf.float32,
        shape=  [None, 20, width])  # [bsz,seq,width]

    # uncomment for masked CNN
    #"""
    state_PH = tf.placeholder(
        name=   'state_PH',
        dtype=  tf.float32,
        shape=  [None, n_lays, kernel-1, width])  # [bsz,n_lay,2,width]
    #"""
    #state_PH = None

    encoder = enc_CNN(
        input=      input_PH,
        history=    state_PH,
        n_layers=   n_lays,
        kernel=     kernel,
        n_filters=  width,
        verb=       2)




"""

 2018 (c) piteren

    PSDD - Parameters Space Definition Dict
        Dictionary that defines space of parameters, is used as an "simple" representation of space by some classes

        PSDD - {axis(parameter name): list or tuple or value}
            > list of ints or floats defines continuous range
            > tuple may contain elements of any type (even non-numeric), may have one
            > any other type is considered to be a constant (single value)

            example:
            {   'a':    [0.0, 1],               # range of floats
                'b':    (-1,-7,10,15.5,90,30),  # set of num(float) values, num will be sorted
                'c':    ('tat','mam','kot'),    # set of diff values
                'd':    [0,10],                 # range of ints
                'e':    (-2.0,2,None)}          # set of diff values
                'f':    (16.2,)}                # single value

"""

from typing import Any, Dict, List, Tuple

from pypaq.lipytools.little_methods import float_to_str

AXIS =  str                                     # axis type (parameter name)
P_VAL = float or int or Any                     # point value (parameter value)
POINT = Dict[AXIS, P_VAL]                       # point (dictionary {parameter: value}) ~ DNA

RANGE = List[P_VAL] or Tuple[P_VAL] or P_VAL    # axis range type (range of parameter)
PSDD  = Dict[AXIS, RANGE]                       # Parameters Space Definition Dict {parameter: range}}


# prepares nice string of POINT
def point_str(p: POINT) -> str:
    s = '{'
    for axis in sorted(list(p.keys())):
        val = p[axis]
        vs = float_to_str(val) if type(val) is float else str(val)
        s += f'{axis}:{vs} '
    s = s[:-1] + '}'
    return s

# TODO: put here functions that (check, validate, extract info from) PSDD


import numpy as np
from typing import Optional, Callable

from pypaq.R4C.policy_gradients.pg_actor import PGActor
from pypaq.R4C.policy_gradients.actor_critic.tf_based.ac_TF_critic_graph import critic_graph
from pypaq.R4C.helpers import RLException
from pypaq.neuralmess.nemodel import NEModel


class AC_TFCritic(PGActor):

    def __init__(
            self,
            name: str=                      'AC_TFCritic',
            nngraph: Optional[Callable]=    critic_graph,
            **kwargs):
        PGActor.__init__(
            self,
            name=       name,
            nnwrap=     NEModel,
            nngraph=    nngraph,
            **kwargs)

    # Critic does not have policy
    def get_policy_probs(self, observation: object) -> np.ndarray:
        raise RLException('not implemented since should not be called')

    def get_qvs(self, observation) -> np.ndarray:
        obs_vec = self._get_observation_vec(observation)
        out = self.nnw(
            feed_dict=  {self.nnw['observation_PH']: [obs_vec]},
            fetch=      ['qvs'])
        return out['qvs']

    def get_qvs_batch(self, observations) -> np.ndarray:
        obs_vecs = self._get_observation_vec_batch(observations)
        out = self.nnw(
            feed_dict=  {self.nnw['observation_PH']: obs_vecs},
            fetch=      ['qvs'])
        return out['qvs']

    def update_with_experience(
            self,
            observations,
            actions_OH,
            next_action_qvs,
            next_actions_probs,
            rewards,
            inspect=    False) -> dict:
        obs_vecs = self._get_observation_vec_batch(observations)
        out = self.nnw.backward(
            feed_dict=  {
                self.nnw['observation_PH']:       obs_vecs,
                self.nnw['action_OH_PH']:         actions_OH,
                self.nnw['next_action_qvs_PH']:   next_action_qvs,
                self.nnw['next_action_probs_PH']: next_actions_probs,
                self.nnw['reward_PH']:            rewards},
            fetch=      ['optimizer','loss','gg_norm','gg_avt_norm'])
        out.pop('optimizer')
        return out


from typing import List

from pypaq.neuralmess.base_elements import tf
from pypaq.neuralmess.layers import lay_dense, zeroes


def pga_graph(
        name=               'pg_actor_graph',
        observation_width=  4,
        num_actions=        2,
        hidden_layers=      (20,),
        lay_norm=           False,
        seed=               121):

    with tf.variable_scope(name):

        observation_PH = tf.placeholder(
            shape=  (None, observation_width),
            dtype=  tf.float32,
            name=   'observation')
        action_PH = tf.placeholder(
            shape=  None,
            dtype=  tf.int32,
            name=   'action')
        return_PH = tf.placeholder( # discounted accumulated return
            shape=  None,
            dtype=  tf.float32,
            name=   'return')

        zsL: List[tf.Tensor] = []  # list of zeroes Tensors

        layer = observation_PH
        if lay_norm: layer = tf.keras.layers.LayerNormalization(axis=-1)(layer)
        for i in range(len(hidden_layers)):
            layer = lay_dense(
                input=      layer,
                name=       f'hidden_layer_{i}',
                units=      hidden_layers[i],
                activation= tf.nn.relu,
                seed=       seed)
            zsL.append(zeroes(layer))
            if lay_norm: layer = tf.keras.layers.LayerNormalization(axis=-1)(layer)

        action_logits = lay_dense(
            input=      layer,
            name=       'action_logits',
            units=      num_actions,
            activation= None,
            seed=       seed)
        probs = tf.nn.softmax(action_logits)

        # actor_ce = tf.losses.sparse_softmax_cross_entropy(logits=action_logits, labels=action_PH, weights=return_PH) # alternate version already weighted
        actor_ce = tf.nn.sparse_softmax_cross_entropy_with_logits(logits=action_logits, labels=action_PH)
        actor_ce_mean = tf.reduce_mean(actor_ce)
        loss = tf.reduce_mean(return_PH * actor_ce) # return * policy(a|s)

    return {
        'observation_PH':   observation_PH,
        'return_PH':        return_PH,
        'action_PH':        action_PH,
        'probs':            probs,
        'actor_ce_mean':    actor_ce_mean,
        'loss':             loss,
        'zeroes':           zsL}


"""

 2022 (c) piteren

    DQN_Actor - Deep QLearningActor NN based

"""

from abc import ABC
import numpy as np
from typing import Optional, Union, Callable, List

from pypaq.R4C.qlearning.ql_actor import QLearningActor
from pypaq.R4C.helpers import RLException
from pypaq.comoneural.nnwrap import NNWrap


# DQN (NN based) QLearningActor
class DQN_Actor(QLearningActor, ABC):

    def __init__(
            self,
            nnwrap: type(NNWrap),
            nngraph: Optional[Union[Callable, type]]=   None,
            **kwargs):

        QLearningActor.__init__(self, **kwargs)

        # some overrides and updates
        kwargs['name'] = self.name
        kwargs['name_timestamp'] = False                # name timestamp is driven by TrainableActor (with self.name)
        if 'logger' in kwargs: kwargs.pop('logger')     # NNWrap will always create own logger (since then it is not given) with optionally given level
        kwargs['num_actions'] = self._envy.num_actions()
        kwargs['observation_width'] = self._get_observation_vec(self._envy.get_observation()).shape[-1]

        self.nnw: NNWrap = nnwrap(nngraph=nngraph, **kwargs)

        self._rlog.info('*** DQN_Actor *** initialized')
        self._rlog.info(f'> NNWrap: {nnwrap.__name__}')

    # vectorization of observations batch, may be overridden with more optimal custom implementation
    def _get_observation_vec_batch(self, observations: List[object]) -> np.ndarray:
        return np.array([self._get_observation_vec(v) for v in observations])

    # INFO: wont be used since DQN_Actor updates only with batches
    def _upd_QV(
            self,
            observation: object,
            action: int,
            new_qv: float) -> float:
        raise RLException('not implemented')

    def _get_save_topdir(self) -> str:
        return self.nnw['save_topdir']

    def save(self):
        self.nnw.save()

    def __str__(self) -> str:
        return str(self.nnw)


"""
    Configuration Manager
        - keeps configuration POINT {key: value}
        - loads from and saves to file
"""
from copy import deepcopy
from typing import Optional

from pypaq.lipytools.files import r_json, w_json
from pypaq.pms.base_types import POINT


class ConfigManager:

    def __init__(
            self,
            file: str,
            config: Optional[POINT]=    None,   # {param: value}
            try_to_load=                True):  # tries to load from file if file exists

        self.__file = file
        self.__config: POINT = config if config is not None else {}
        if try_to_load: self.load()
        self.__save_file()

    # loads configuration from file and returns it
    def __read_file(self) -> POINT:
        return r_json(self.__file)

    # saves configuration to jsonl file
    def __save_file(self):
        w_json(self.__config, self.__file)

    # returns (copy of) config
    def get_config(self) -> POINT:
        return deepcopy(self.__config)

    # loads configuration from file, updates self, returns new configuration (from file) or only keys that have changed values
    def load(
            self,
            return_only_changed=    True) -> POINT:

        file_config = self.__read_file() or {}
        config_changed = {}
        for k in file_config:
            if k not in self.__config or self.__config[k] != file_config[k]:
                self.__config[k] = file_config[k]
                config_changed[k] = file_config[k]

        return config_changed if return_only_changed else file_config

    # updates self with given kwargs, saves file if needed
    def update(
            self,
            return_only_changed=    True,
            **kwargs) -> POINT:

        config_changed = {}
        for k in kwargs:
            if k not in self.__config:
                self.__config[k] = kwargs[k]
                config_changed[k] = kwargs[k]
            else:
                if kwargs[k] != self.__config[k]:
                    self.__config[k] = kwargs[k]
                    config_changed[k] = kwargs[k]

        if config_changed: self.__save_file()

        return config_changed if return_only_changed else self.get_config()


def test_config(n_loops=10):

    import time

    config = {
        'param_aaa':    15,
        'beta':         20.45,
        'do_it':        False,
        'dont_do':      None}

    cm = ConfigManager(
        config= config,
        file=   'config.file')
    print(cm.get_config())

    for _ in range(n_loops):
        time.sleep(5)
        newc = cm.load()
        print(newc, cm.get_config())


if __name__ == '__main__':
    test_config()



import torch
from typing import Optional

from pypaq.torchness.motorch import Module
from pypaq.torchness.layers import LayDense, zeroes
from pypaq.torchness.base_elements import scaled_cross_entropy


class A2CModel(Module):

    def __init__(
            self,
            observation_width: int=             4,
            num_actions: int=                   2,
            two_towers: bool=                   False,  # builds separate towers for Actor & Critic
            num_layers: int=                    1,
            layer_width: int=                   50,
            lay_norm: bool=                     False,
            clamp_advantage: Optional[float]=   0.5,    # limits advantage abs value
            use_scaled_ce: bool=                True,   # experimental Scaled Cross Entropy loss
            use_huber: bool=                    False,  # for True uses Huber loss for Critic
            opt_class=                          torch.optim.Adam,
            #opt_class=                          torch.optim.SGD,
            #opt_momentum=                       0.5,
            #opt_nesterov=                       True,
            # RMSProp, Adadelta
    ):

        torch.nn.Module.__init__(self)

        hidden_layers = tuple([layer_width] * num_layers)

        lay_shapeL = []
        next_in = observation_width
        for hl in hidden_layers:
            lay_shapeL.append((next_in,hl))
            next_in = hl

        self.lay_norm = lay_norm

        self.ln = torch.nn.LayerNorm(observation_width) # input layer norm

        self.linL = [LayDense(*shape) for shape in lay_shapeL]
        self.lnL = [torch.nn.LayerNorm(shape[-1]) for shape in lay_shapeL]
        lix = 0
        for lin,ln in zip(self.linL, self.lnL):
            self.add_module(f'lay_lin{lix}', lin)
            self.add_module(f'lay_ln{lix}', ln)
            lix += 1

        self.linL_tower = [LayDense(*shape) for shape in lay_shapeL] if two_towers else []
        self.lnL_tower = [torch.nn.LayerNorm(shape[-1]) for shape in lay_shapeL] if two_towers else []
        lix = 0
        for lin,ln in zip(self.linL_tower, self.lnL_tower):
            self.add_module(f'lay_lin_tower{lix}', lin)
            self.add_module(f'lay_ln_tower{lix}', ln)
            lix += 1

        self.value = LayDense(
            in_features=    next_in,
            out_features=   1,
            activation=     None)

        self.logits = LayDense(
            in_features=    next_in,
            out_features=   num_actions,
            activation=     None)

        self.clamp_advantage = clamp_advantage
        self.use_scaled_ce = use_scaled_ce
        self.use_huber = use_huber

    def forward(self, observation) -> dict:

        inp = self.ln(observation) if self.lay_norm else observation

        zsL = []
        out = inp
        for lin,ln in zip(self.linL,self.lnL):
            out = lin(out)
            zsL.append(zeroes(out))
            if self.lay_norm: out = ln(out)

        out_tower = out
        if self.linL_tower:
            out_tower = inp
            for lin,ln in zip(self.linL_tower,self.lnL_tower):
                out_tower = lin(out_tower)
                zsL.append(zeroes(out_tower))
                if self.lay_norm: out_tower = ln(out_tower)
        value = self.value(out_tower)
        value = torch.reshape(value, (value.shape[:-1])) # remove last dim

        logits = self.logits(out)
        probs = torch.nn.functional.softmax(input=logits, dim=-1)

        return {
            'value':    value,
            'logits':   logits,
            'probs':    probs,
            'zeroes':   zsL}

    def loss_acc(self, observation, action_taken, dreturn) -> dict:

        out = self(observation)

        value = out['value']
        logits = out['logits']

        advantage = dreturn - value

        advantage_nograd = advantage.detach() # to prevent flow of Actor loss gradients to Critic network
        if self.clamp_advantage is not None:
            advantage_nograd = torch.clamp(
                input=  advantage_nograd,
                min=    -self.clamp_advantage,
                max=    self.clamp_advantage)

        if self.use_scaled_ce:
            actor_ce_scaled = scaled_cross_entropy(
                labels= action_taken,
                scale=  advantage_nograd,
                probs=  out['probs'])
        else:
            actor_ce = torch.nn.functional.cross_entropy(logits, action_taken, reduction='none')
            actor_ce_scaled = actor_ce * advantage_nograd

        loss_actor_scaled_mean = torch.mean(actor_ce_scaled)

        if self.use_huber: loss_critic = torch.nn.functional.huber_loss(value, dreturn, reduction='none')
        else:              loss_critic = advantage * advantage  # MSE

        loss_critic_mean = torch.mean(loss_critic)

        out.update({
            'advantage':    advantage,
            'loss_actor':   loss_actor_scaled_mean,
            'loss_critic':  loss_critic_mean,
            'loss':         loss_actor_scaled_mean + loss_critic_mean})

        return out


# to suppress TF logs from C side by setting an environment variable
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import numpy as np
import tensorflow as tf
from typing import List, Callable, Dict, Optional
import warnings

from pypaq.lipytools.little_methods import stamp, get_params, get_func_dna
from pypaq.lipytools.logger import set_logger
from pypaq.lipytools.moving_average import MovAvg
from pypaq.mpython.devices import get_devices
from pypaq.pms.parasave import ParaSave
from pypaq.neuralmess_duo.base_elements import lr_scaler, grad_clipper_AVT
from pypaq.neuralmess_duo.tbwr import TBwr
from pypaq.comoneural.batcher import Batcher

# restricted keys for fwd_func DNA and return DNA (if they appear in kwargs, should be named exactly like below)
SPEC_KEYS = [
    'name',         # model name
    'train_vars',   # list of variables to train (may be returned, otherwise all trainable are taken)
    'opt_vars',     # list of variables returned by opt_func
    'loss',         # loss
    'acc',          # accuracy
    'f1',           # F1
    'batch_size',   # batch size
    'n_batches',    # number of batches for train
    'verb',         # verbosity
]

# defaults below may be given with NEModelDUO kwargs or overridden by fwd_graph attributes
NEMODELDUO_DEFAULTS = {
    'seed':         123,                        # seed for TF and numpy
    'opt_class':    tf.keras.optimizers.Adam,   # default optimizer of train()
    'devices':      'GPU:0',                    # for TF1 we used '/device:CPU:0'
    # LR management (parameters of LR warmup and annealing)
    'baseLR':          3e-4,                       # initial learning rate (base init)
    'warm_up':      None,
    'ann_base':     None,
    'ann_step':     1.0,
    'n_wup_off':    1.0,
    # gradients clipping parameters
    'avt_SVal':     0.1,
    'avt_window':   100,
    'avt_max_upd':  1.5,
    'do_clip':      False,
    # other
    'save_topdir':  '_models',                  # top folder of model save
    'save_fn_pfx':  'nemodelduo_dna',           # dna filename prefix
    'hpmser_mode':  False,                      # it will set model to be read_only and quiet when running with hpmser
    'read_only':    False,                      # sets model to be read only - wont save anything (wont even create self.nnwrap_dir)
    'do_logfile':   True,                       # enables saving log file in self.nnwrap_dir
    'do_TB':        True}                       # runs TensorBard, saves in self.nnwrap_dir


# exemplary FWD function implementation
def fwd_graph(
        in_width=           10,
        hidden_layers=      (128,),
        out_width=          3,
        baseLR=                0.0005,     # defaults of fwd_graph will override NEMODELDUO_DEFAULTS
        verb=               0):

    in_vec =  tf.keras.Input(shape=in_width, name="in_vec")
    in_true = tf.keras.Input(shape=1,        name="in_true", dtype=tf.int32)
    if verb>0: print(' > in_vec', in_vec)

    lay = in_vec
    for i in range(len(hidden_layers)):
        lay = tf.keras.layers.Dense(
            name=       f'hidden_layer_{i}',
            units=      hidden_layers[i],
            activation= tf.nn.relu)(lay)

    logits = tf.keras.layers.Dense(
        name=       'value',
        units=      out_width,
        activation= None)(lay)
    if verb>0: print(' > logits', logits)

    probs = tf.nn.softmax(logits)

    loss = tf.keras.losses.sparse_categorical_crossentropy(
        y_true=         in_true,
        y_pred=         logits,
        from_logits=    True)
    loss = tf.reduce_mean(loss)
    if verb > 0: print(' > loss', loss)

    preds = tf.cast(tf.argmax(logits, axis=-1), dtype=tf.int32)
    acc = tf.reduce_mean(tf.cast(tf.equal(tf.squeeze(in_true), preds), dtype=tf.float32))

    return {
        'in_vec':   in_vec,
        'in_true':  in_true,
        'probs':    probs,
        'loss':     loss, # loss must be returned to build train_model
        'acc':      acc,

        # train_model IO specs, put here inputs and output (keys) of train_model to be built by NEModelDUO
        'train_model_IO':   {
            'inputs':       ['in_vec', 'in_true'],  # str or List[str]
            'outputs':      ['probs', 'acc']},      # str or List[str], put here any metric tensors, loss may be not added
    }



class NEModelDUO(ParaSave):

    def __init__(
            self,
            name: str,
            fwd_func: Optional[Callable]=   None,       # function building graph (from inputs to loss) - may be not given if model already saved
            name_timestamp=                 False,      # adds timestamp to the model name
            save_topdir=                    NEMODELDUO_DEFAULTS['save_topdir'],
            save_fn_pfx=                    NEMODELDUO_DEFAULTS['save_fn_pfx'],
            verb=                           0,
            **kwargs):

        if not tf.executing_eagerly(): warnings.warn(f'TF is NOT executing eagerly!')

        if name_timestamp: name += f'.{stamp()}'
        if verb>0: print(f'\n *** NEModelDUO {name} (type: {type(self).__name__}) *** initializes..')

        # ***************************************************************************************** manage (resolve) DNA

        # load dna from folder
        dna_saved = ParaSave.load_dna(
            name=           name,
            save_topdir=    save_topdir,
            save_fn_pfx=    save_fn_pfx)

        if not fwd_func:
            assert 'fwd_func' in dna_saved, 'ERR: cannot continue: fwd_func was not given and model ahs not been saved before, please give fwd_func to model init!'
            fwd_func = dna_saved['fwd_func']

        dna_fwd_func_defaults = get_params(fwd_func)['with_defaults']  # get fwd_func defaults

        dna = {}
        dna.update(NEMODELDUO_DEFAULTS)
        dna.update(dna_fwd_func_defaults)
        dna.update(dna_saved)
        dna.update(kwargs)                  # update with kwargs given NOW by user
        dna.update({
            'name':         name,
            'save_topdir':  save_topdir,
            'save_fn_pfx':  save_fn_pfx,
            'fwd_func':     fwd_func,
            'verb':         verb})

        ParaSave.__init__(self, lock_managed_params=True, **dna)
        self.check_params_sim(SPEC_KEYS + list(NEMODELDUO_DEFAULTS.keys())) # safety check

        # hpmser_mode - early override
        if self['hpmser_mode']:
            self.verb = 0
            self['read_only'] = True

        # read only - early override
        if self['read_only']:
            self['do_logfile'] = False
            self['do_TB'] = False

        self.nnwrap_dir = f'{self.save_topdir}/{self.name}'
        if self.verb>0: print(f' > NEModelDUO dir: {self.nnwrap_dir}{" read only mode!" if self["read_only"] else ""}')

        if self['do_logfile']:
            set_logger(
                log_folder=     self.nnwrap_dir,
                custom_name=    self.name,
                verb=           self.verb)

        dna = self.get_point()  # bake

        if self.verb>0:
            print(f'\n > NEModelDUO DNA sources:')
            print(f' >> NEMODELDUO_DEFAULTS:    {NEMODELDUO_DEFAULTS}')
            print(f' >> fwd_func defaults:      {dna_fwd_func_defaults}')
            print(f' >> DNA saved:              {dna_saved}')
            print(f' >> given kwargs:           {kwargs}')
            print(f' NEModelDUO complete DNA:   {dna}')

        # **************************************************************************************************************

        np.random.seed(self['seed'])
        tf.random.set_seed(self['seed'])

        # *********************************************************************************************** manage devices

        # TODO: by now we are using only first device, support for multidevice will be added later
        self.device = get_devices(self['devices'], namespace='TF2')[0]
        if self.verb>1: print(f' > setting devices from {self["devices"]} to {self.device}')

        # check for TF visible_physical_devices
        visible_physical_devices = tf.config.list_physical_devices()
        if self.verb>1:
            print(' > devices available for TF:')
            for dev in visible_physical_devices:
                print(f' >> {dev.name}')

        # INFO: Currently, memory growth needs to be the same across GPUs (https://www.tensorflow.org/guide/gpu#logging_device_placement)
        for dev in visible_physical_devices:
            if 'GPU' in dev.name:
                tf.config.experimental.set_memory_growth(dev, True)

        set_visible = []
        for dev in visible_physical_devices:
            # INFO: it looks that TF usually needs CPU and it cannot be masked-out
            # TODO: self.device in dev.name will fail for more than 10 GPUs, cause 'GPU:1' is in 'GPU:10' - need to fix it later
            if self.device in dev.name or 'CPU' in dev.name:
                set_visible.append(dev)
        if self.verb>1: print(f' > setting visible to NEModelDUO: {set_visible} for {self.device}')
        tf.config.set_visible_devices(set_visible)

        # **************************************************************************************************************

        if self.verb>0: print(f'\n > building graph ({fwd_func}) on {self.device} ..')
        if self.verb>1: tf.debugging.set_log_device_placement(True)

        fwd_func_dna = get_func_dna(fwd_func, dna)
        with tf.device(self.device):
            fwd_func_out = fwd_func(**fwd_func_dna)
        self.update(fwd_func_out)

        assert 'loss' in self, 'ERR: You need to return loss with fwd_func!'
        assert 'train_model_IO' in self, 'ERR: fwd_func should return train_model_IO specs, see fwd_graph example!'

        # change to lists
        for k in self['train_model_IO']:
            if type(self['train_model_IO'][k]) is str:
                self['train_model_IO'][k] = [self['train_model_IO'][k]]
        if 'outputs' not in self['train_model_IO']: self['train_model_IO']['outputs'] = []
        if 'loss' not in self['train_model_IO']['outputs']: self['train_model_IO']['outputs'].append('loss') # add loss

        if self.verb>0: print(f'\n > NEModelDUO is building train_model, name: {self.name} ..')
        self.train_model = self.__get_model(
            name=       self.name,
            inputs=     self['train_model_IO']['inputs'],
            outputs=    self['train_model_IO']['outputs'])

        with tf.device(self.device):

            # variable for time averaged global norm of gradients
            self.lr = self.train_model.add_weight(
                name=       'lr',
                trainable=  False,
                dtype=      tf.float32)
            self.lr.assign(self['baseLR'])

            # variable for time averaged global norm of gradients
            self.ggnorm_avt = self.train_model.add_weight(
                name=       'gg_avt_norm',
                trainable=  False,
                dtype=      tf.float32)
            self.ggnorm_avt.assign(self['avt_SVal'])

            # global step variable
            self.iterations = self.train_model.add_weight(
                name=       'iterations',
                trainable=  False,
                dtype=      tf.int64)
            self.iterations.assign(0)

        try:
            self.train_model.load_weights(filepath=f'{self.nnwrap_dir}/weights')
            if self.verb>0: print(f' > train_model weights loaded..')
        except Exception as e:
            if self.verb>0: print(f' > train_model weights NOT loaded ({e})..')

        if self.verb>1:
            print(f'\n >> train.model ({self.train_model.name}) weights:')
            for w in self.train_model.weights: print(f' **  {w.name:30} {str(w.shape):10} {w.device}')

        scaled_LR = lr_scaler(
            baseLR=        self.lr,
            g_step=     self.iterations,
            warm_up=    self['warm_up'],
            ann_base=   self['ann_base'],
            ann_step=   self['ann_step'],
            n_wup_off=  self['n_wup_off'],
            verb=       self.verb)
        if self.verb>0: print('scaled_LR', scaled_LR)
        self.optimizer = self['opt_class'](learning_rate=scaled_LR)
        self.optimizer.iterations = self.iterations

        self.submodels: Dict[str, tf.keras.Model] = {}

        self.writer = TBwr(logdir=self.nnwrap_dir, set_to_CPU=False) if self['do_TB'] else None

        self._model_data = None
        self._batcher = None

        if self.verb>0: print(f'\n > NEModelDUO init finished..')


    def __get_model(
            self,
            name: str,
            inputs: str or List[str],
            outputs: str or List[str]) -> tf.keras.Model:
        if type(inputs) is str:  inputs = [inputs]
        if type(outputs) is str: outputs = [outputs]
        for n in inputs + outputs: assert n in self, f'ERR: {n} not in self!'
        return tf.keras.Model(
            name=       name,
            inputs=     {n: self[n] for n in inputs},
            outputs=    {n: self[n] for n in outputs})


    def build_callable_submodel(
            self,
            name: str,
            inputs: str or List[str],
            outputs: str or List[str]):
        assert name not in self.submodels
        if self.verb>0: print(f'\n > NEModelDUO is building callable: {name}, inputs: {inputs}, outputs: {outputs}')
        self.submodels[name] = self.__get_model(
            name=       name,
            inputs=     inputs,
            outputs=    outputs)

    # call wrapped with tf.function
    @tf.function(reduce_retracing=True)
    def __call(self, data, name):

        if name is None: model = self.train_model
        else:
            assert name in self.submodels
            model = self.submodels[name]

        if self.verb > 1: print(
            f' >> NEModelDUO is calling: {model.name}, inputs: {model.inputs}, outputs: {model.outputs}')
        return model(data, training=False)

    # call wrapped with device
    def call(self, data, name: Optional[str]=None):
        with tf.device(self.device):
            return self.__call(data, name)

    # train wrapped with tf.function
    @tf.function(reduce_retracing=True)
    def __train_batch(self, data):

        with tf.GradientTape() as tape:
            out = self.train_model(data, training=True)

        # TODO: what about colocate_gradients_with_ops=False
        gradients = tape.gradient(
            target=     out['loss'],
            sources=    self.train_model.trainable_variables)

        gclr_out = grad_clipper_AVT(
            variables=      self.train_model.trainable_variables,
            gradients=      gradients,
            ggnorm_avt=     self.ggnorm_avt,
            optimizer=      self['optimizer'],
            avt_window=     self['avt_window'],
            avt_max_upd=    self['avt_max_upd'],
            do_clip=        self['do_clip'],
            verb=           self.verb)

        out.update({
            'ggnorm':       gclr_out['ggnorm'],
            'ggnorm_avt':   self.ggnorm_avt,
            'iterations':   self['optimizer'].iterations})

        return out

    # __train_batch wrapped with device
    def train_batch(self, data):
        with tf.device(self.device):
            return self.__train_batch(data)

    # ******************************************************************** baseline methods for training with given data

    # loads model data for training, dict should have at least 'train':{} for Batcher
    def load_model_data(self) -> dict:
        warnings.warn('NEModelBase.load_model_data() should be overridden!')
        return {}

    # pre (before) training method - may be overridden
    def pre_train(self):
        self._model_data = self.load_model_data()
        self._batcher = Batcher(
            data_TR=        self._model_data['train'],
            data_VL=        self._model_data['valid'] if 'valid' in self._model_data else None,
            data_TS=        self._model_data['test'] if 'test' in self._model_data else None,
            batch_size=     self['batch_size'],
            batching_type=  'random_cov',
            verb=           self.verb)

    # training method, saves for max_ts_acc
    # INFO: training method below is based on model accuracy so it should be returned by graph as 'acc'
    def train(
            self,
            test_freq=      100,  # number of batches between tests, model SHOULD BE tested while training
            mov_avg_factor= 0.1,
            save=           True):  # allows to save model while training

        self.pre_train()

        if self.verb>0: print(f'{self.name} - training starts')
        batch_IX = 0
        tr_lssL = []
        tr_accL = []
        ts_acc_max = 0
        ts_acc_mav = MovAvg(mov_avg_factor)

        ts_results = []
        ts_bIX = [bIX for bIX in range(self['n_batches']+1) if not bIX % test_freq] # batch indexes when test will be performed
        assert ts_bIX, 'ERR: model SHOULD BE tested while training!'
        ten_factor = int(0.1*len(ts_bIX)) # number of tests for last 10% of training
        if ten_factor < 1: ten_factor = 1 # we need at least one result
        if self['hpmser_mode']: ts_bIX = ts_bIX[-ten_factor:]

        while batch_IX < self['n_batches']:
            batch_IX += 1
            batch = self._batcher.get_batch()

            run_out = self.train_batch(batch)

            if self['do_TB']:
                for k in ['loss', 'acc', 'ggnorm', 'ggnorm_avt']:
                    self.log_TB(value=run_out[k], tag=f'tr/{k}', step=batch_IX)

            if self.verb>0:
                tr_lssL.append(run_out['loss'])
                tr_accL.append(run_out['acc'])

            if batch_IX in ts_bIX:
                ts_out = self.test()
                ts_results.append(ts_out['acc'])
                ts_out['acc_mav'] = ts_acc_mav.upd(ts_out['acc'])
                if self['do_TB']:
                    for k in ['loss', 'acc', 'acc_mav']:
                        self.log_TB(value=ts_out[k], tag=f'ts/{k}', step=batch_IX)
                if self.verb>0: print(f'{batch_IX:5d} TR: {100*sum(tr_accL)/test_freq:.1f} / {sum(tr_lssL)/test_freq:.3f} -- TS: {100*ts_out["acc"]:.1f} / {ts_out["loss"]:.3f}')
                tr_lssL = []
                tr_accL = []

                if ts_out['acc'] > ts_acc_max:
                    ts_acc_max = ts_out['acc']
                    if not self['read_only'] and save: self.save()  # model is saved for max_ts_acc

        # weighted test value for last 10% test results
        ts_results = ts_results[-ten_factor:]
        ts_wval = 0
        weight = 1
        sum_weight = 0
        for tr in ts_results:
            ts_wval += tr * weight
            sum_weight += weight
            weight += 1
        ts_wval /= sum_weight
        if self['do_TB']: self.log_TB(value=ts_wval, tag='ts/ts_wval', step=batch_IX)
        if self.verb>0:
            print(f'model {self.name} finished training')
            print(f' > test_acc_max: {ts_acc_max:.4f}')
            print(f' > test_wval:    {ts_wval:.4f}')

        return ts_wval


    def test(self):
        batches = self._batcher.get_TS_batches()
        lossL = []
        accL = []
        for batch in batches:
            out = self.train_model(batch, training=False)
            lossL.append(out['loss'])
            accL.append(out['acc'])
        return {
            'acc':  sum(accL)/len(accL),
            'loss': sum(lossL)/len(lossL)}


    def log_TB(self, value, tag: str, step: int):
        if self.writer: self.writer.add(value=value, tag=tag, step=step)
        else: warnings.warn(f'NEModelDUO {self.name} cannot log TensorBoard since do_TB flag is False!')

    # updates base LR (baseLR) in graph - but not saves it to the checkpoint
    def update_LR(self, lr: float):
        old = self['baseLR']
        self['baseLR'] = lr
        self.lr.assign(self['baseLR'])
        if self.verb>1: print(f'NEModelDUO {self.name} updated baseLR from {old} to {self["baseLR"]}')


    def __str__(self):
        s = f'NEModelDUO {self.name}'
        s += f'\n > baseLR: {self.lr.numpy()}'
        s += f'\n > iterations: {self.iterations.numpy()}'
        s += f'\ntrain_model weights:'
        total_w = 0
        for w in self.train_model.weights:
            s += f'\n > {w.name:30} {str(w.shape):10} {w.device}'
            shape = w.shape
            variable_parameters = 1
            for dim in shape:
                variable_parameters *= dim
            total_w += variable_parameters
        s += f'\n ### total variables: {total_w}'
        s += f'\nModel params:\n{ParaSave.dict_2str(self.get_point())}'
        return s


    def save(self):
        assert not self['read_only'], 'ERR: read only NEModelDUO cannot be saved!'
        self.save_dna()
        self.iterations.assign(self['optimizer'].iterations)
        self.train_model.save_weights(filepath=f'{self.nnwrap_dir}/weights')
        if self.verb>0: print(f'model {self.name} saved')


    def exit(self):
        if self.writer: self.writer.exit()

    # copies full NEModelDUO folder (DNA & checkpoints)
    @staticmethod
    def copy_saved(
            name_src: str,
            name_trg: str,
            save_topdir_src: str=           NEMODELDUO_DEFAULTS['save_topdir'],
            save_topdir_trg: Optional[str]= None,
            save_fn_pfx: str=               NEMODELDUO_DEFAULTS['save_fn_pfx']):

        if save_topdir_trg is None: save_topdir_trg = save_topdir_src

        # copy DNA with ParaSave
        ParaSave.copy_saved_dna(
            name_src=           name_src,
            name_trg=           name_trg,
            save_topdir_src=    save_topdir_src,
            save_topdir_trg=    save_topdir_trg,
            save_fn_pfx=        save_fn_pfx)

        # load dna from folder & build child model
        dna_trg = ParaSave.load_dna(
            name=           name_trg,
            save_topdir=    save_topdir_trg,
            save_fn_pfx=    save_fn_pfx)
        model_trg = NEModelDUO(**dna_trg)

        ckptA_FD = f'{save_topdir_src}/{name_src}/'
        model_trg.train_model.load_weights(filepath=f'{ckptA_FD}/weights')

        ckptC_FD = f'{save_topdir_trg}/{name_trg}/'
        model_trg.train_model.save_weights(filepath=f'{ckptC_FD}/weights')

        model_trg.exit()

    # returns train_model weights (returns list of numpy.ndarray, numpy.float32, numpy.int64, ..)
    def get_weights(self) -> list:
        return [w.numpy() for w in self.train_model.weights]

    @staticmethod
    def gx_saved(
            name_parent_main: str,
            name_parent_scnd: Optional[str],            # if not given makes GX only with main parent
            name_child: str,
            save_topdir_parent_main: str=               NEMODELDUO_DEFAULTS['save_topdir'],
            save_topdir_parent_scnd: Optional[str] =    None,
            save_topdir_child: Optional[str] =          None,
            save_fn_pfx: Optional[str] =                NEMODELDUO_DEFAULTS['save_fn_pfx'],
            ratio: float=                               0.5,
            noise: float=                               0.03) -> None:

        if not save_topdir_parent_scnd: save_topdir_parent_scnd = save_topdir_parent_main
        if not save_topdir_child: save_topdir_child = save_topdir_parent_main

        # GX ParaSave dna
        ParaSave.gx_saved_dna(
            name_parent_main=           name_parent_main,
            name_parent_scnd=           name_parent_scnd,
            name_child=                 name_child,
            save_topdir_parent_main=    save_topdir_parent_main,
            save_topdir_parent_scnd=    save_topdir_parent_scnd,
            save_topdir_child=          save_topdir_child,
            save_fn_pfx=                save_fn_pfx)

        # load dna from folder & build child model
        dna_saved = ParaSave.load_dna(
            name=           name_child,
            save_topdir=    save_topdir_child,
            save_fn_pfx=    save_fn_pfx)
        model_child = NEModelDUO(**dna_saved)

        ckptA_FD = f'{save_topdir_parent_main}/{name_parent_main}/'
        model_child.train_model.load_weights(filepath=f'{ckptA_FD}/weights')
        weights_A = model_child.get_weights()

        if name_parent_scnd:
            ckptB_FD = f'{save_topdir_parent_scnd}/{name_parent_scnd}/'
            model_child.train_model.load_weights(filepath=f'{ckptB_FD}/weights')

        for w,a in zip(model_child.train_model.weights, weights_A):
            if np.issubdtype(a.dtype, np.floating):
                noise_tensor = tf.random.truncated_normal(
                    shape=  w.shape,
                    stddev= tf.math.reduce_std(w))
                new_val = w * (1 - ratio) + a * ratio + noise_tensor * noise
                w.assign(new_val)

        ckptC_FD = f'{save_topdir_child}/{name_child}/'
        model_child.train_model.save_weights(filepath=f'{ckptC_FD}/weights')

        model_child.exit()

        """
        # other unsuccessful tryouts to load wieghts from TF2 checkpoint:
        
        ckpt = tf.train.load_checkpoint(ckptA_FD)
        print(ckpt)
        print(tf.train.list_variables(ckptA_FD))
        
        ckpt = tf.train.Checkpoint()
        print(ckpt)
        ckpt.restore(tf.train.latest_checkpoint(ckptA_FD))
        print(ckpt)
        
        lsv = tf.train.list_variables(tf.train.latest_checkpoint(ckptA_FD))
        print(lsv)
        for e in lsv:
            v = tf.train.load_variable(ckptA_FD, e[0])
            print(type(v))
        """